from typing import Generator, Iterable, List, Optional, Sequence, Tuple, Union

import math

import numpy

from functools import reduce
from itertools import chain
from numbers import Number

from numpy import ndarray
from numpy.ma import MaskedArray
from vtk import vtkImageData, vtkXMLImageDataWriter
from vtk.util import numpy_support

from . import util, errors

basestring = (str, bytes)


def silence_numpy_errors():
    """Silence numpy warnings about zero division. Normal usage of Space() will trigger these warnings."""
    numpy.seterr(divide="ignore", invalid="ignore")


def sum_onto(a, axis):
    """Numpy convenience. Project all dimensions of an array onto an axis, i.e. apply sum() to all axes except the one given."""
    for i in reversed(list(range(len(a.shape)))):
        if axis != i:
            a = a.sum(axis=i)
    return a


class Axis(object):
    """Represents a single dimension finite discrete grid centered at 0.

    Important attributes:
    min     lower bound
    max     upper bound
    res     step size / resolution
    label   human-readable identifier

    min, max and res are floats, but internally only integer operations are used. In particular
    min = imin * res, max = imax * res
    """

    def __init__(self, min, max, res, label=None):
        self.res = float(res)
        if isinstance(min, int):
            self.imin = min
        else:
            self.imin = math.floor(min / self.res)
        if isinstance(max, int):
            self.imax = max
        else:
            self.imax = math.ceil(max / self.res)
        self.label = label

    @property
    def max(self):
        return self.imax * self.res

    @property
    def min(self):
        return self.imin * self.res

    def __len__(self):
        return self.imax - self.imin + 1

    def __iter__(self):
        return iter(self[index] for index in range(len(self)))

    def __getitem__(self, key):
        if isinstance(key, slice):
            if key.step is not None:
                raise IndexError("stride not supported")
            if key.start is None:
                start = 0
            elif key.start < 0:
                raise IndexError("key out of range")
            elif isinstance(key.start, int):
                start = key.start
            else:
                raise IndexError("key start must be integer")
            if key.stop is None:
                stop = len(self)
            elif key.stop > len(self):
                raise IndexError("key out of range")
            elif isinstance(key.stop, int):
                stop = key.stop
            else:
                raise IndexError("slice stop must be integer")
            return self.__class__(
                self.imin + start, self.imin + stop - 1, self.res, self.label
            )
        elif isinstance(key, int) or isinstance(key, numpy.int64):
            if key >= len(self):  # to support iteration
                raise IndexError("key out of range")
            return (self.imin + key) * self.res
        else:
            raise IndexError("unknown key {0!r}".format(key))

    def _get_index_float(self, value: float) -> int:
        intvalue = int(round(value / self.res))
        if self.imin <= intvalue <= self.imax:
            return intvalue - self.imin
        raise ValueError(
            "cannot get index: value {0} not in range [{1}, {2}]".format(
                value, self.min, self.max
            )
        )

    def get_index(self,
                  value: Union[float, slice, ndarray]) -> Union[int, slice, ndarray]:
        if isinstance(value, (int, float)):  # nice mypy accept int as float but not isinstance
            return self._get_index_float(value)
        elif isinstance(value, slice):
            if value.step is not None:
                raise IndexError("stride not supported")
            if value.start is None:
                start = None
            else:
                start = self._get_index_float(value.start)
            if value.stop is None:
                stop = None
            else:
                stop = self._get_index_float(value.stop)
            if start is not None and stop is not None and start > stop:
                start, stop = stop, start
            return slice(start, stop)
        else:  # TODO it seems that we have maskedarray here...
            intvalue = numpy.around(value / self.res).astype(int)
            if ((self.imin <= intvalue) & (intvalue <= self.imax)).all():
                return intvalue - self.imin
            raise ValueError(
                "cannot get indices, values from [{0}, {1}], axes range [{2}, {3}]".format(
                    value.min(), value.max(), self.min, self.max
                )
            )

    def __or__(self, other):  # union operation
        if not isinstance(other, Axis):
            return NotImplemented
        if not self.is_compatible(other):
            raise ValueError("cannot unite axes with different resolution/label")
        return self.__class__(
            min(self.imin, other.imin), max(self.imax, other.imax), self.res, self.label
        )

    def __eq__(self, other):
        if not isinstance(other, Axis):
            return NotImplemented
        return (
            self.res == other.res
            and self.imin == other.imin
            and self.imax == other.imax
            and self.label == other.label
        )

    def __hash__(self):
        return hash(self.imin) ^ hash(self.imax) ^ hash(self.res) ^ hash(self.label)

    def is_compatible(self, other):
        if not isinstance(other, Axis):
            return False
        return self.res == other.res and self.label == other.label

    def __contains__(self, other):
        if isinstance(other, Number):
            return self.min <= other <= self.max
        elif isinstance(other, Axis):
            return (
                self.is_compatible(other)
                and self.imin <= other.imin
                and self.imax >= other.imax
            )

    def rebound(self, min, max):
        return self.__class__(min, max, self.res, self.label)

    def rebin(self, factor):
        # for integers the following relations hold: a // b == floor(a / b), -(-a // b) == ceil(a / b)
        new = self.__class__(
            self.imin // factor, -(-self.imax // factor), factor * self.res, self.label
        )
        return self.imin % factor, -self.imax % factor, new

    def __repr__(self):
        return "{0.__class__.__name__} {0.label} (min={0.min}, max={0.max}, res={0.res}, count={1})".format(
            self, len(self)
        )

    def restrict_slice(self, value: slice) -> slice:
        # TODO rename into clip_slice
        if value.step is not None:
            raise IndexError("stride not supported")

        v = value.start # type: Optional[float]
        if value.start is None:
            start = None
        else:
            start = numpy.clip(value.start, self.min, self.max)

        if value.stop is None:
            stop = None
        else:
            stop = numpy.clip(value.stop, self.min, self.max)

        # TODO this produce a wrong slice sementic, the stop value can not be None.
        # if value.stop == self.max:
        #     stop = None

        if start is not None and stop is not None and start > stop:
            start, stop = stop, start

        # TODO par of the code necessite that start and stop are not None.
        # this is not the usual slice semantic
        assert start is not None, f"1-{start}"
        assert stop is not None, f"2-{stop}"

        return slice(start, stop)

    def restrict(self, value: Union[float, slice]) -> Union[float, slice]:
        if isinstance(value, float):
            return numpy.clip(value, self.min, self.max)
        elif isinstance(value, slice):
            return self.restrict_slice(value)


class Axes(object):
    """Luxurious tuple of Axis objects."""

    def __init__(self, axes):
        self.axes = tuple(axes)
        if len(self.axes) > 1 and any(axis.label is None for axis in self.axes):
            raise ValueError("axis label is required for multidimensional space")

    def __iter__(self):
        return iter(self.axes)

    @property
    def dimension(self):
        return len(self.axes)

    @property
    def npoints(self):
        return numpy.array([len(ax) for ax in self.axes]).prod()

    @property
    def memory_size(self):
        # assuming double precision floats for photons, 32 bit integers for contributions
        return (8 + 4) * self.npoints

    @classmethod
    def fromfile(cls, filename: str) -> 'Axes':
        with util.open_h5py(filename, "r") as fp:
            try:
                if "axes" in fp and "axes_labels" in fp:
                    # oldest style, float min/max
                    return cls(
                        tuple(
                            Axis(min, max, res, lbl)
                            for ((min, max, res), lbl) in zip(
                                fp["axes"], fp["axes_labels"]
                            )
                        )
                    )
                elif "axes" in fp:  # new
                    try:
                        axes = tuple(
                            Axis(int(imin), int(imax), res, lbl)
                            for ((index, fmin, fmax, res, imin, imax), lbl) in zip(
                                fp["axes"].values(), fp["axes"].keys()
                            )
                        )
                        return cls(
                            tuple(
                                axes[int(values[0])] for values in fp["axes"].values()
                            )
                        )  # reorder the axes to the way in which they were saved
                    except ValueError:
                        return cls(
                            tuple(
                                Axis(int(imin), int(imax), res, lbl)
                                for ((imin, imax, res), lbl) in zip(
                                    fp["axes"].values(), fp["axes"].keys()
                                )
                            )
                        )
                else:
                    # older style, integer min/max
                    return cls(
                        tuple(
                            Axis(imin, imax, res, lbl)
                            for ((imin, imax), res, lbl) in zip(
                                fp["axes_range"], fp["axes_res"], fp["axes_labels"]
                            )
                        )
                    )
            except (KeyError, TypeError) as e:
                raise errors.HDF5FileError(
                    "unable to load axes definition from HDF5 file {0}, is it a valid BINoculars file? (original error: {1!r})".format(
                        filename, e
                    )
                )

    def tofile(self, filename):
        with util.open_h5py(filename, "w") as fp:
            axes = fp.create_group("axes")
            for index, ax in enumerate(self.axes):
                axes.create_dataset(
                    ax.label, data=[index, ax.min, ax.max, ax.res, ax.imin, ax.imax]
                )

    def toarray(self):
        return numpy.vstack(
            numpy.hstack([str(ax.imin), str(ax.imax), str(ax.res), ax.label])
            for ax in self.axes
        )

    @classmethod
    def fromarray(cls, arr):
        return cls(
            tuple(
                Axis(int(imin), int(imax), float(res), lbl)
                for (imin, imax, res, lbl) in arr
            )
        )

    def index(self, obj: Union[Axis, int, str]) -> int:
        if isinstance(obj, Axis):
            return self.axes.index(obj)
        elif isinstance(obj, int) and 0 <= obj < len(self.axes):
            return obj
        elif isinstance(obj, basestring):
            label = obj.lower()
            matches = tuple(
                i for i, axis in enumerate(self.axes) if axis.label.lower() == label
            )
            if len(matches) == 0:
                raise ValueError("no matching axis found")
            elif len(matches) == 1:
                return matches[0]
            else:
                raise ValueError("ambiguous axis label {0}".format(label))
        else:
            raise ValueError("invalid axis identifier {0!r}".format(obj))

    def __contains__(self, obj):
        if isinstance(obj, Axis):
            return obj in self.axes
        elif isinstance(obj, int):
            return 0 <= obj < len(self.axes)
        elif isinstance(obj, basestring):
            label = obj.lower()
            return any(axis.label.lower() == label for axis in self.axes)
        else:
            raise ValueError("invalid axis identifier {0!r}".format(obj))

    def __len__(self):
        return len(self.axes)

    def __getitem__(self, key):
        return self.axes[key]

    def __eq__(self, other):
        if not isinstance(other, Axes):
            return NotImplemented
        return self.axes == other.axes

    def __ne__(self, other):
        if not isinstance(other, Axes):
            return NotImplemented
        return self.axes != other.axes

    def __repr__(self):
        return "{0.__class__.__name__} ({0.dimension} dimensions, {0.npoints} points, {1}) {{\n    {2}\n}}".format(
            self,
            util.format_bytes(self.memory_size),
            "\n    ".join(repr(ax) for ax in self.axes),
        )

    def restricted_key(self, key):
        if len(key) == 0:
            return None
        if len(key) == len(self.axes):
            return tuple(ax.restrict(s) for s, ax in zip(key, self.axes))
        else:
            raise IndexError("dimension mismatch")


class EmptySpace(object):
    """Convenience object for sum() and friends. Treated as zero for addition.
    Does not share a base class with Space for simplicity."""

    def __init__(self, config=None, metadata=None):
        self.config = config
        self.metadata = metadata

    def __add__(self, other):
        if not isinstance(other, Space) and not isinstance(other, EmptySpace):
            return NotImplemented
        return other

    def __radd__(self, other):
        if not isinstance(other, Space) and not isinstance(other, EmptySpace):
            return NotImplemented
        return other

    def __iadd__(self, other):
        if not isinstance(other, Space) and not isinstance(other, EmptySpace):
            return NotImplemented
        return other

    def tofile(self, filename):
        """Store EmptySpace in HDF5 file."""
        with util.atomic_write(filename) as tmpname:
            with util.open_h5py(tmpname, "w") as fp:
                fp.attrs["type"] = "Empty"

    def __repr__(self):
        return "{0.__class__.__name__}".format(self)


class Space(object):
    """Main data-storing object in BINoculars.
    Data is represented on an n-dimensional rectangular grid. Per grid point,
    the number of photons (~ intensity) and the number of original data points
    (pixels) contribution is stored.

    Important attributes:
        axes             Axes instances describing range and stepsizes of each of the dimensions
        photons          n-dimension numpy float array, total intensity per grid point
        contribitions    n-dimensional numpy integer array, number of original datapoints (pixels) per grid point
        dimension        n"""

    def __init__(self, axes, config=None, metadata=None):
        if not isinstance(axes, Axes):
            self.axes = Axes(axes)
        else:
            self.axes = axes

        self.config = config
        self.metadata = metadata

        self.photons = numpy.zeros([len(ax) for ax in self.axes], order="C")
        self.contributions = numpy.zeros(self.photons.shape, order="C")

    @property
    def dimension(self):
        return self.axes.dimension

    @property
    def npoints(self):
        return self.photons.size

    @property
    def memory_size(self):
        """Returns approximate memory consumption of this Space.
        Only considers size of .photons and .contributions, does not take into account the overhead."""
        return self.photons.nbytes + self.contributions.nbytes

    @property
    def config(self):
        """util.ConfigFile instance describing configuration file used to create this Space instance"""
        return self._config

    @config.setter
    def config(self, conf):
        if isinstance(conf, util.ConfigFile):
            self._config = conf
        elif not conf:
            self._config = util.ConfigFile()
        else:
            raise TypeError("'{0!r}' is not a util.ConfigFile".format(conf))

    @property
    def metadata(self):
        """util.MetaData instance describing metadata used to create this Space instance"""
        return self._metadata

    @metadata.setter
    def metadata(self, metadata):
        if isinstance(metadata, util.MetaData):
            self._metadata = metadata
        elif not metadata:
            self._metadata = util.MetaData()
        else:
            raise TypeError("'{0!r}' is not a util.MetaData".format(metadata))

    def copy(self):
        """Returns a copy of self. Numpy data is not shared, but the Axes object is."""
        new = self.__class__(self.axes, self.config, self.metadata)
        new.photons[:] = self.photons
        new.contributions[:] = self.contributions
        return new

    def get(self):
        """Returns normalized photon count."""
        return self.photons / self.contributions

    def __repr__(self):
        return "{0.__class__.__name__} ({0.dimension} dimensions, {0.npoints} points, {1}) {{\n    {2}\n}}".format(
            self,
            util.format_bytes(self.memory_size),
            "\n    ".join(repr(ax) for ax in self.axes),
        )

    def __getitem__(self, key):
        """Slicing only! space[-0.2:0.2, 0.9:1.1] does exactly what the syntax implies.
        Ellipsis operator '...' is not supported."""

        newkey = self.get_key(key)
        newaxes = tuple(
            ax[k] for k, ax in zip(newkey, self.axes) if isinstance(ax[k], Axis)
        )
        if not newaxes:
            return self.photons[newkey] / self.contributions[newkey]
        newspace = self.__class__(newaxes, self.config, self.metadata)
        newspace.photons = self.photons[newkey].copy()
        newspace.contributions = self.contributions[newkey].copy()
        return newspace

    def get_key(self,
        key: Union[float,
                   slice,
                   Tuple[Union[float, slice, ndarray], ...],
                   List[Union[float, slice, ndarray]]]
        ) -> Tuple[Union[int, slice, ndarray], ...]:
        # needed in the fitaid for visualising the interpolated data
        """Convert the n-dimensional interval described by key (as used by e.g. __getitem__()) from data coordinates to indices."""
        if isinstance(key, (int, float)) or isinstance(key, slice):
            if not len(self.axes) == 1:
                raise IndexError("dimension mismatch")
            else:
                key = [key]
        elif not (isinstance(key, tuple) or isinstance(key, list)) or not len(
            key
        ) == len(self.axes):
            raise IndexError("dimension mismatch")
        return tuple(ax.get_index(k) for k, ax in zip(key, self.axes))

    def project(self, axis: Union[str, int],
                *more_axes) -> 'Space':
        """Reduce dimensionality of Space by projecting onto 'axis'.
        All data (photons, contributions) is summed along this axis.

        axis         the label of the axis or the index
        *more_axis   also project on these axes"""

        index = self.axes.index(axis)
        newaxes = list(self.axes)
        newaxes.pop(index)
        newspace = self.__class__(newaxes, self.config, self.metadata)
        newspace.photons = self.photons.sum(axis=index)
        newspace.contributions = self.contributions.sum(axis=index)
        if more_axes:
            return newspace.project(more_axes[0], *more_axes[1:])
        else:
            return newspace

    def slice(self, axis, key):
        """Single-axis slice.

        axis    label or index of axis to slice
        key     something like slice(lower_data_range, upper_data_range)"""
        axindex = self.axes.index(axis)
        newkey = list(slice(None) for ax in self.axes)
        newkey[axindex] = key
        return self.__getitem__(tuple(newkey))

    def get_masked(self) -> MaskedArray:
        """Returns photons/contributions, but with divide-by-zero's masked out."""
        return numpy.ma.array(data=self.get(), mask=(self.contributions == 0))

    def get_variance(self):
        return numpy.ma.array(
            data=1 / self.contributions, mask=(self.contributions == 0)
        )

    def get_grid(self) -> Tuple[ndarray]:
        """Returns the data coordinates of each grid point, as n-tuple of n-dimensinonal arrays.
        Basically numpy.mgrid() in data coordinates."""
        igrid = numpy.mgrid[tuple(slice(0, len(ax)) for ax in self.axes)]
        grid = tuple(
            numpy.array((grid + ax.imin) * ax.res) for grid, ax in zip(igrid, self.axes)
        )
        return grid

    def max(self, axis=None):
        """Returns maximum intensity."""
        return self.get_masked().max(axis=axis)

    def argmax(self):
        """Returns data coordinates of grid point with maximum intensity."""
        array = self.get_masked()
        return tuple(
            ax[key]
            for ax, key in zip(
                self.axes, numpy.unravel_index(numpy.argmax(array), array.shape)
            )
        )

    def __add__(self, other):
        if isinstance(other, Number):
            new = self.copy()
            new.photons += other * self.contributions
            return new
        if not isinstance(other, Space):
            return NotImplemented
        if not len(self.axes) == len(other.axes) or not all(
            a.is_compatible(b) for (a, b) in zip(self.axes, other.axes)
        ):
            raise ValueError(
                "cannot add spaces with different dimensionality or resolution"
            )

        new = self.__class__([a | b for (a, b) in zip(self.axes, other.axes)])
        new += self
        new += other
        return new

    def __iadd__(self, other):
        if isinstance(other, Number):
            self.photons += other * self.contributions
            return self
        if not isinstance(other, Space):
            return NotImplemented
        if not len(self.axes) == len(other.axes) or not all(
            a.is_compatible(b) for (a, b) in zip(self.axes, other.axes)
        ):
            raise ValueError(
                "cannot add spaces with different dimensionality or resolution"
            )

        if not all(
            other_ax in self_ax for (self_ax, other_ax) in zip(self.axes, other.axes)
        ):
            return self.__add__(other)

        index = tuple(
            slice(
                self_ax.get_index(other_ax.min),
                self_ax.get_index(other_ax.min) + len(other_ax),
            )
            for (self_ax, other_ax) in zip(self.axes, other.axes)
        )
        self.photons[index] += other.photons
        self.contributions[index] += other.contributions
        self.metadata += other.metadata
        return self

    def __sub__(self, other):
        return self.__add__(other * -1)

    def __isub__(self, other):
        return self.__iadd__(other * -1)

    def __mul__(self, other):
        if isinstance(other, Number):
            new = self.__class__(self.axes, self.config, self.metadata)
            # we would like to keep 1/contributions as the variance
            # var(aX) = a**2var(X)
            new.photons = self.photons / other
            new.contributions = self.contributions / other ** 2
            return new
        else:
            return NotImplemented

    def trim(self):
        """Reduce total size of Space by trimming zero-contribution data points on the boundaries."""
        mask = self.contributions > 0
        lims = (
            numpy.flatnonzero(sum_onto(mask, i)) for (i, ax) in enumerate(self.axes)
        )
        lims = tuple((i.min(), i.max()) for i in lims)
        self.axes = Axes(
            ax.rebound(min + ax.imin, max + ax.imin)
            for (ax, (min, max)) in zip(self.axes, lims)
        )
        slices = tuple(slice(min, max + 1) for (min, max) in lims)
        self.photons = self.photons[slices].copy()
        self.contributions = self.contributions[slices].copy()

    def rebin(self, resolutions):
        """Change bin size.

        resolution    n-tuple of floats, new resolution of each axis"""
        if not len(resolutions) == len(self.axes):
            raise ValueError("cannot rebin space with different dimensionality")
        if resolutions == tuple(ax.res for ax in self.axes):
            return self

        # gather data and transform
        labels = list(ax.label for ax in self.axes)
        coords = self.get_grid()
        intensity = self.get()
        weights = self.contributions
        return self.from_image(resolutions, labels, coords, intensity, weights)

    def reorder(self, labels):
        """Change order of axes."""
        if not self.dimension == len(labels):
            raise ValueError("dimension mismatch")
        newindices = list(self.axes.index(label) for label in labels)
        new = self.__class__(
            tuple(self.axes[index] for index in newindices), self.config, self.metadata
        )
        new.photons = numpy.transpose(self.photons, axes=newindices)
        new.contributions = numpy.transpose(self.contributions, axes=newindices)
        return new

    def transform_coordinates(self, resolutions, labels, transformation):
        # gather data and transform
        coords = self.get_grid()
        transcoords = transformation(*coords)
        intensity = self.get()
        weights = self.contributions

        # get rid of invalid coords
        valid = reduce(
            numpy.bitwise_and,
            chain((numpy.isfinite(t) for t in transcoords)),
            (weights > 0),
        )
        transcoords = tuple(t[valid] for t in transcoords)

        return self.from_image(
            resolutions, labels, transcoords, intensity[valid], weights[valid]
        )

    def process_image(self, coordinates, intensity, weights):
        """Load image data into Space.

        coordinates  n-tuple of data coordinate arrays
        intensity    data intensity array
        weights      weights array, supply numpy.ones_like(intensity) for equal weights"""
        if len(coordinates) != len(self.axes):
            raise ValueError("dimension mismatch between coordinates and axes")

        intensity = numpy.nan_to_num(
            intensity
        ).flatten()  # invalids can be handeled by setting weight to 0, this ensures the weights can do that
        weights = weights.flatten()

        indices = numpy.array(
            tuple(ax.get_index(coord) for (ax, coord) in zip(self.axes, coordinates))
        )
        for i in range(0, len(self.axes)):
            for j in range(i + 1, len(self.axes)):
                indices[i, :] *= len(self.axes[j])
        indices = indices.sum(axis=0).astype(int).flatten()

        photons = numpy.bincount(indices, weights=intensity * weights)
        contributions = numpy.bincount(indices, weights=weights)

        self.photons.ravel()[: photons.size] += photons
        self.contributions.ravel()[: contributions.size] += contributions

    @classmethod
    def from_image(
        cls, resolutions, labels, coordinates, intensity, weights, limits=None
    ):
        """Create Space from image data.

        resolutions   n-tuple of axis resolutions
        labels          n-tuple of axis labels
        coordinates   n-tuple of data coordinate arrays
        intensity     data intensity array"""

        if limits is not None:
            invalid = numpy.zeros(intensity.shape).astype(numpy.bool)
            for coord, sl in zip(coordinates, limits):
                if sl.start is None and sl.stop is not None:
                    invalid += coord > sl.stop
                elif sl.start is not None and sl.stop is None:
                    invalid += coord < sl.start
                elif sl.start is not None and sl.stop is not None:
                    invalid += numpy.bitwise_or(coord < sl.start, coord > sl.stop)

            if numpy.all(invalid == True):
                return EmptySpace()
            coordinates = tuple(coord[~invalid] for coord in coordinates)
            intensity = intensity[~invalid]
            weights = weights[~invalid]

        axes = tuple(
            Axis(coord.min(), coord.max(), res, label)
            for res, label, coord in zip(resolutions, labels, coordinates)
        )
        newspace = cls(axes)
        newspace.process_image(coordinates, intensity, weights)
        return newspace

    def tovti(self, filename: str) -> None:
        assert self.dimension == 3, "only array with three dimensions are allowed"

        data = self.photons / self.contributions

        spacing = tuple(a.res for a in self.axes.axes)
        origin = tuple(a.min for a in self.axes.axes)
        name = str(tuple(a.label for a in self.axes.axes))

        image_data = vtkImageData()
        image_data.SetSpacing(spacing)
        image_data.SetOrigin(origin)
        image_data.SetDimensions(data.shape)

        temp_array = numpy.transpose(numpy.flip(data, 2)).flatten()
        temp_array = numpy_support.numpy_to_vtk(temp_array, deep=1)

        pd = image_data.GetPointData()
        pd.SetScalars(temp_array)
        pd.GetArray(0).SetName(name)

        # export data to file
        writer = vtkXMLImageDataWriter()
        writer.SetFileName(filename)
        writer.SetInputData(image_data)
        writer.Write()

    def tofile(self, filename):
        """Store Space in HDF5 file."""
        with util.atomic_write(filename) as tmpname:
            with util.open_h5py(tmpname, "w") as fp:
                fp.attrs["type"] = "Space"
                self.config.tofile(fp)
                self.axes.tofile(fp)
                self.metadata.tofile(fp)
                fp.create_dataset(
                    "counts",
                    self.photons.shape,
                    dtype=self.photons.dtype,
                    compression="gzip",
                ).write_direct(self.photons)
                fp.create_dataset(
                    "contributions",
                    self.contributions.shape,
                    dtype=self.contributions.dtype,
                    compression="gzip",
                ).write_direct(self.contributions)

    @classmethod
    def fromfile(cls, file: str, key: Optional[Sequence[slice]]=None) -> 'Space':
        """Load Space from HDF5 file.

        file      filename string or h5py.Group instance
        key       sliced (subset) loading, should be an n-tuple of slice()s in data coordinates"""
        try:
            with util.open_h5py(file, "r") as fp:
                if "type" in fp.attrs.keys():
                    if fp.attrs["type"] == "Empty":
                        return EmptySpace()

                axes = Axes.fromfile(fp)
                config = util.ConfigFile.fromfile(fp)
                metadata = util.MetaData.fromfile(fp)
                if key:
                    if len(axes) != len(key):
                        raise ValueError(
                            "dimensionality of 'key' does not match dimensionality of Space in HDF5 file {0}".format(
                                file
                            )
                        )
                    key = tuple(ax.get_index(k) for k, ax in zip(key, axes))
                    for index, sl in enumerate(key):
                        if sl.start == sl.stop and sl.start is not None:
                            raise KeyError("key results in empty space")
                    axes = tuple(
                        ax[k] for k, ax in zip(key, axes) if isinstance(k, slice)
                    )
                else:
                    key = Ellipsis
                space = cls(axes, config, metadata)
                try:
                    fp["counts"].read_direct(space.photons, key)
                    fp["contributions"].read_direct(space.contributions, key)
                except (KeyError, TypeError) as e:
                    raise errors.HDF5FileError(
                        "unable to load Space from HDF5 file {0}, is it a valid BINoculars file? (original error: {1!r})".format(
                            file, e
                        )
                    )

        except IOError as e:
            raise errors.HDF5FileError(
                "unable to open '{0}' as HDF5 file (original error: {1!r})".format(
                    file, e
                )
            )
        return space


class Multiverse(object):
    """A collection of spaces with basic support for addition.
       Only to be used when processing data. This makes it possible to
       process multiple limit sets in a combination of scans"""

    def __init__(self, spaces):
        self.spaces = list(spaces)

    @property
    def dimension(self):
        return len(self.spaces)

    def __add__(self, other):
        if not isinstance(other, Multiverse):
            return NotImplemented
        if not self.dimension == other.dimension:
            raise ValueError("cannot add multiverses with different dimensionality")
        return self.__class__(tuple(s + o for s, o in zip(self.spaces, other.spaces)))

    def __iadd__(self, other):
        if not isinstance(other, Multiverse):
            return NotImplemented
        if not self.dimension == other.dimension:
            raise ValueError("cannot add multiverses with different dimensionality")
        for index, o in enumerate(other.spaces):
            self.spaces[index] += o
        return self

    def tofile(self, filename):
        with util.atomic_write(filename) as tmpname:
            with util.open_h5py(tmpname, "w") as fp:
                fp.attrs["type"] = "Multiverse"
                for index, sp in enumerate(self.spaces):
                    spacegroup = fp.create_group("space_{0}".format(index))
                    sp.tofile(spacegroup)

    @classmethod
    def fromfile(cls, file):
        """Load Multiverse from HDF5 file."""
        try:
            with util.open_h5py(file, "r") as fp:
                if "type" in fp.attrs:
                    if fp.attrs["type"] == "Multiverse":
                        return cls(tuple(Space.fromfile(fp[label]) for label in fp))
                    else:
                        raise TypeError("This is not a multiverse")
                else:
                    raise TypeError("This is not a multiverse")
        except IOError as e:
            raise errors.HDF5FileError(
                "unable to open '{0}' as HDF5 file (original error: {1!r})".format(
                    file, e
                )
            )

    def __repr__(self):
        return "{0.__class__.__name__}\n{1}".format(self, self.spaces)


class EmptyVerse(object):
    """Convenience object for sum() and friends. Treated as zero for addition."""

    spaces = [EmptySpace()]

    @property
    def dimension(self):
        return len(self.spaces)

    def __add__(self, other):
        if not isinstance(other, Multiverse):
            return NotImplemented
        return other

    def __radd__(self, other):
        if not isinstance(other, Multiverse):
            return NotImplemented
        return other

    def __iadd__(self, other):
        if not isinstance(other, Multiverse):
            return NotImplemented
        return other


def union_axes(axes):
    axes = tuple(axes)
    if len(axes) == 1:
        return axes[0]
    if not all(isinstance(ax, Axis) for ax in axes):
        raise TypeError("not all objects are Axis instances")
    if len(set(ax.res for ax in axes)) != 1 or len(set(ax.label for ax in axes)) != 1:
        raise ValueError("cannot unite axes with different resolution/label")
    mi = min(ax.min for ax in axes)
    ma = max(ax.max for ax in axes)
    first = axes[0]
    return first.__class__(mi, ma, first.res, first.label)


def union_unequal_axes(axes):
    axes = tuple(axes)
    if len(axes) == 1:
        return axes[0]
    if not all(isinstance(ax, Axis) for ax in axes):
        raise TypeError("not all objects are Axis instances")
    if len(set(ax.label for ax in axes)) != 1:
        raise ValueError("cannot unite axes with different label")
    mi = min(ax.min for ax in axes)
    ma = max(ax.max for ax in axes)
    res = min(
        ax.res for ax in axes
    )  # making it easier to use the sliderwidget otherwise this hase no meaning
    first = axes[0]
    return first.__class__(mi, ma, res, first.label)


def sum(spaces):
    """Calculate sum of iterable of Space instances."""
    spaces = tuple(space for space in spaces if not isinstance(space, EmptySpace))
    if len(spaces) == 0:
        return EmptySpace()
    if len(spaces) == 1:
        return spaces[0]
    if len(set(space.dimension for space in spaces)) != 1:
        raise TypeError("dimension mismatch in spaces")

    first = spaces[0]
    axes = tuple(
        union_axes(space.axes[i] for space in spaces) for i in range(first.dimension)
    )
    newspace = first.__class__(axes)
    for space in spaces:
        newspace += space
    return newspace


def verse_sum(verses: Generator[Multiverse, None, None]) -> Multiverse:
    i = iter(M.spaces for M in verses)
    return Multiverse(sum(spaces) for spaces in zip(*i))


# hybrid sum() / __iadd__()


def chunked_sum(verses: Iterable[Multiverse], chunksize=10) -> Union[EmptyVerse, Multiverse]:
    """Calculate sum of iterable of Multiverse instances. Creates intermediate sums to avoid growing a large space at every summation.

    verses     iterable of Multiverse instances
    chunksize  number of Multiverse instances in each intermediate sum"""
    result = EmptyVerse()
    for chunk in util.grouper(verses, chunksize):
        result += verse_sum(M for M in chunk)
    return result


def iterate_over_axis(space, axis, resolution=None):
    ax = space.axes[space.axes.index(axis)]
    if resolution:
        bins = get_bins(ax, resolution)
        for start, stop in zip(bins[:-1], bins[1:]):
            yield space.slice(axis, slice(start, stop))
    else:
        for value in ax:
            yield space.slice(axis, value)


def get_axis_values(axes, axis, resolution=None) -> ndarray:
    ax = axes[axes.index(axis)]
    if resolution:
        bins = get_bins(ax, resolution)
        return (bins[:-1] + bins[1:]) / 2
    else:
        return numpy.array(list(ax))


def iterate_over_axis_keys(axes, axis, resolution=None):
    axindex = axes.index(axis)
    ax = axes[axindex]
    k = [slice(None) for i in axes]
    if resolution:
        bins = get_bins(ax, resolution)
        for start, stop in zip(bins[:-1], bins[1:]):
            k[axindex] = slice(start, stop)
            yield k
    else:
        for value in ax:
            k[axindex] = value
            yield k


def get_bins(ax: Axis, resolution: float) -> ndarray:
    if float(resolution) < ax.res:
        raise ValueError(
            "interval {0} to low, minimum interval is {1}".format(resolution, ax.res)
        )

    mi, ma = ax.min, ax.max
    if(resolution >= (ma - mi)):
        return numpy.array([mi, ma])
    else:
        return numpy.linspace(mi, ma, math.ceil(1.0 / resolution * (ma - mi)))


def dstack(spaces, dindices, dlabel, dresolution):
    def transform(space, dindex):
        resolutions = list(ax.res for ax in space.axes)
        resolutions.append(dresolution)
        labels = list(ax.label for ax in space.axes)
        labels.append(dlabel)
        exprs = list(ax.label for ax in space.axes)
        exprs.append("ones_like({0}) * {1}".format(labels[0], dindex))
        transformation = util.transformation_from_expressions(space, exprs)
        return space.transform_coordinates(resolutions, labels, transformation)

    return sum(transform(space, dindex) for space, dindex in zip(spaces, dindices))


def axis_offset(space, label, offset):
    exprs = list(ax.label for ax in space.axes)
    index = space.axes.index(label)
    exprs[index] += "+ {0}".format(offset)
    transformation = util.transformation_from_expressions(space, exprs)
    return space.transform_coordinates(
        (ax.res for ax in space.axes), (ax.label for ax in space.axes), transformation
    )


def bkgsubtract(space, bkg):
    if space.dimension == bkg.dimension:
        bkg.photons = bkg.photons * space.contributions / bkg.contributions
        bkg.photons[bkg.contributions == 0] = 0
        bkg.contributions = space.contributions
        return space - bkg
    else:
        photons = numpy.broadcast_arrays(space.photons, bkg.photons)[1]
        contributions = numpy.broadcast_arrays(space.contributions, bkg.contributions)[
            1
        ]
        bkg = Space(space.axes)
        bkg.photons = photons
        bkg.contributions = contributions
        return bkgsubtract(space, bkg)


def make_compatible(spaces):
    if not numpy.alen(numpy.unique(len(space.axes) for space in spaces)) == 1:
        raise ValueError("cannot make spaces with different dimensionality compatible")
    ax0 = tuple(ax.label for ax in spaces[0].axes)
    resmax = tuple(
        numpy.vstack(
            tuple(ax.res for ax in space.reorder(ax0).axes) for space in spaces
        ).max(axis=0)
    )
    resmin = tuple(
        numpy.vstack(
            tuple(ax.res for ax in space.reorder(ax0).axes) for space in spaces
        ).min(axis=0)
    )
    if not resmax == resmin:
        print(
            "Warning: Not all spaces have the same resolution. Resolution will be changed to: {0}".format(
                resmax
            )
        )
    return tuple(space.reorder(ax0).rebin2(resmax) for space in spaces)



from binoculars.backends import id03
import binoculars.util
import binoculars.space
import os
import numpy

import unittest


class TestCase(unittest.TestCase):
    def setUp(self):
        cfg_unparsed = {}
        specfile = os.path.join(
            os.path.split(os.getcwd())[0],
            "binoculars-binaries/examples/dataset/sixc_tutorial.spec",
        )
        cfg_unparsed["specfile"] = specfile
        cfg_unparsed["sdd"] = "1000"
        cfg_unparsed["pixelsize"] = "0.055, 0.055"
        cfg_unparsed["imagefolder"] = specfile.replace("sixc_tutorial.spec", "images")
        cfg_unparsed["centralpixel"] = "50 ,50"
        numpy.save("mask.npy", numpy.identity(516))
        cfg_unparsed["maskmatrix"] = "mask.npy"
        self.id03input = id03.EH2(cfg_unparsed)
        self.projection = id03.HKLProjection(
            {"resolution": "0.01", "limits": "[0:, :-1, 0:0.2]"}
        )

    @unittest.expectedFailure
    def test_IO(self):
        jobs = list(self.id03input.generate_jobs(["820"]))
        destination_opts = self.id03input.get_destination_options(["820"])
        imagedata = self.id03input.process_job(jobs[0])
        intensity, weights, coords = imagedata.next()
        projected = self.projection.project(*coords)
        limits = self.projection.config.limits

        space1 = binoculars.space.Space.from_image(
            self.projection.config.resolution,
            self.projection.get_axis_labels(),
            projected,
            intensity,
            weights,
            limits=limits[0],
        )

        print(space1)
        intensity, weights, coords = imagedata.next()
        projected = self.projection.project(*coords)

        space2 = binoculars.space.Space.from_image(
            self.projection.config.resolution,
            self.projection.get_axis_labels(),
            projected,
            intensity,
            weights,
        )
        print(space1 + space2)

    def tearDown(self):
        os.remove("mask.npy")


if __name__ == "__main__":
    unittest.main()



import numpy
import scipy.optimize
import scipy.special
import inspect
import re


class FitBase(object):
    parameters = None
    guess = None
    result = None
    summary = None
    fitdata = None

    def __init__(self, space, guess=None):
        self.space = space
        code = inspect.getsource(self.func)

        args = tuple(
            re.findall("\((.*?)\)", line)[0].split(",")
            for line in code.split("\n")[2:4]
        )

        if space.dimension != len(args[0]):
            raise ValueError(
                "dimension mismatch: space has {0}, {1.__class__.__name__} expects {2}".format(
                    space.dimension, self, len(args[0])
                )
            )
        self.parameters = args[1]

        self.xdata, self.ydata, self.cxdata, self.cydata = self._prepare(self.space)
        if guess is not None:
            if len(guess) != len(self.parameters):
                raise ValueError(
                    "invalid number of guess parameters {0!r} for {1!r}".format(
                        guess, self.parameters
                    )
                )
            self.guess = guess
        else:
            self._guess()
        self.success = self._fit()

    @staticmethod
    def _prepare(space):
        ydata = space.get_masked()
        cydata = ydata.compressed()
        imask = ~ydata.mask
        xdata = space.get_grid()
        cxdata = tuple(d[imask] for d in xdata)
        return xdata, ydata, cxdata, cydata

    def _guess(self):
        # the implementation should use space and/or self.xdata/self.ydata and/or the cxdata/cydata maskless versions to obtain guess
        raise NotImplementedError

    def _fitfunc(self, params):
        return self.cydata - self.func(self.cxdata, params)

    def _fit(self):
        result = scipy.optimize.leastsq(
            self._fitfunc, self.guess, full_output=True, epsfcn=0.000001
        )

        self.message = re.sub("\s{2,}", " ", result[3].strip())
        self.result = result[0]
        errdata = result[2]["fvec"]
        if result[1] is None:
            self.variance = numpy.zeros(len(self.result))
        else:
            self.variance = numpy.diagonal(
                result[1] * (errdata ** 2).sum() / (len(errdata) - len(self.result))
            )

        self.fitdata = numpy.ma.array(
            self.func(self.xdata, self.result), mask=self.ydata.mask
        )
        self.summary = "\n".join(
            "%s: %.4g +/- %.4g" % (n, p, v)
            for (n, p, v) in zip(self.parameters, self.result, self.variance)
        )

        return result[4] in (
            1,
            2,
            3,
            4,
        )  # corresponds to True on success, False on failure

    def __str__(self):
        return "{0.__class__.__name__} fit on {1}\n{2}\n{3}".format(
            self, self.space, self.message, self.summary
        )


class PeakFitBase(FitBase):
    def __init__(self, space, guess=None, loc=None):
        if loc != None:
            self.argmax = tuple(loc)
        else:
            self.argmax = None
        super(PeakFitBase, self).__init__(space, guess)

    def _guess(self):
        maximum = self.cydata.max()  # for background determination

        background = self.cydata < (numpy.median(self.cydata) + maximum) / 2

        if any(background == True):  # the fit will fail if background is flas for all
            linparams = self._linfit(
                list(grid[background] for grid in self.cxdata), self.cydata[background]
            )
        else:
            linparams = numpy.zeros(len(self.cxdata) + 1)

        simbackground = linparams[-1] + numpy.sum(
            numpy.vstack(
                [
                    param * grid.flatten()
                    for (param, grid) in zip(linparams[:-1], self.cxdata)
                ]
            ),
            axis=0,
        )
        signal = self.cydata - simbackground

        if self.argmax != None:
            argmax = self.argmax
        else:
            argmax = tuple((signal * grid).sum() / signal.sum() for grid in self.cxdata)

        argmax_bkg = linparams[-1] + numpy.sum(
            numpy.vstack(
                [
                    param * grid.flatten()
                    for (param, grid) in zip(linparams[:-1], argmax)
                ]
            )
        )

        try:
            maximum = self.space[argmax] - argmax_bkg
        except ValueError:
            maximum = self.cydata.max()

        if numpy.isnan(maximum):
            maximum = self.cydata.max()

        self.set_guess(maximum, argmax, linparams)

    def _linfit(self, coordinates, intensity):
        coordinates = list(coordinates)
        coordinates.append(numpy.ones_like(coordinates[0]))
        matrix = numpy.vstack([coords.flatten() for coords in coordinates]).T
        return numpy.linalg.lstsq(matrix, intensity, rcond=None)[0]


class AutoDimensionFit(FitBase):
    def __new__(cls, space, guess=None):
        if space.dimension in cls.dimensions:
            return cls.dimensions[space.dimension](space, guess)
        else:
            raise TypeError(
                "{0}-dimensional space not supported for {1.__name__}".format(
                    space.dimension, cls
                )
            )


# utility functions
def rot2d(x, y, th):
    xrot = x * numpy.cos(th) + y * numpy.sin(th)
    yrot = -x * numpy.sin(th) + y * numpy.cos(th)
    return xrot, yrot


def rot3d(x, y, z, th, ph):
    xrot = (
        numpy.cos(th) * x
        + numpy.sin(th) * numpy.sin(ph) * y
        + numpy.sin(th) * numpy.cos(ph) * z
    )
    yrot = numpy.cos(ph) * y - numpy.sin(ph) * z
    zrot = (
        -numpy.sin(th) * x
        + numpy.cos(th) * numpy.sin(ph) * y
        + numpy.cos(th) * numpy.cos(ph) * z
    )
    return xrot, yrot, zrot


def get_class_by_name(name):
    options = {}
    for k, v in globals().items():
        if isinstance(v, type) and issubclass(v, FitBase):
            options[k.lower()] = v
    if name.lower() in options:
        return options[name.lower()]
    else:
        raise ValueError("unsupported fit function '{0}'".format(name))


# fitting functions
class Lorentzian1D(PeakFitBase):
    @staticmethod
    def func(grid, params):
        (x,) = grid
        (I, loc, gamma, slope, offset) = params
        return I / ((x - loc) ** 2 + gamma ** 2) + offset + x * slope

    def set_guess(self, maximum, argmax, linparams):
        gamma0 = 5 * self.space.axes[0].res  # estimated FWHM on 10 pixels
        self.guess = [maximum, argmax[0], gamma0, linparams[0], linparams[1]]


class Lorentzian1DNoBkg(PeakFitBase):
    @staticmethod
    def func(grid, params):
        (x,) = grid
        (I, loc, gamma) = undefined  # TODO
        return I / ((x - loc) ** 2 + gamma ** 2)

    def set_guess(self, maximum, argmax, linparams):
        gamma0 = 5 * self.space.axes[0].res  # estimated FWHM on 10 pixels
        self.guess = [maximum, argmax[0], gamma0]


class PolarLorentzian2Dnobkg(PeakFitBase):
    @staticmethod
    def func(grid, params):
        (x, y) = grid
        (I, loc0, loc1, gamma0, gamma1, th) = params
        a, b = tuple(
            grid - center
            for grid, center in zip(rot2d(x, y, th), rot2d(loc0, loc1, th))
        )
        return I / (1 + (a / gamma0) ** 2 + (b / gamma1) ** 2)

    def set_guess(self, maximum, argmax, linparams):
        gamma0 = self.space.axes[0].res  # estimated FWHM on 10 pixels
        gamma1 = self.space.axes[1].res
        self.guess = [maximum, argmax[0], argmax[1], gamma0, gamma1, 0]


class PolarLorentzian2D(PeakFitBase):
    @staticmethod
    def func(grid, params):
        (x, y) = grid
        (I, loc0, loc1, gamma0, gamma1, th, slope1, slope2, offset) = params
        a, b = tuple(
            grid - center
            for grid, center in zip(rot2d(x, y, th), rot2d(loc0, loc1, th))
        )
        return (
            I / (1 + (a / gamma0) ** 2 + (b / gamma1) ** 2)
            + x * slope1
            + y * slope2
            + offset
        )

    def set_guess(self, maximum, argmax, linparams):
        gamma0 = self.space.axes[0].res  # estimated FWHM on 10 pixels
        gamma1 = self.space.axes[1].res
        self.guess = [
            maximum,
            argmax[0],
            argmax[1],
            gamma0,
            gamma1,
            0,
            linparams[0],
            linparams[1],
            linparams[2],
        ]

    def integrate_signal(self):
        return self.func(
            self.cxdata,
            (
                self.result[0],
                self.result[1],
                self.result[2],
                self.result[3],
                self.result[4],
                self.result[5],
                0,
                0,
                0,
            ),
        ).sum()


class Lorentzian2D(PeakFitBase):
    @staticmethod
    def func(grid, params):
        (x, y) = grid
        (I, loc0, loc1, gamma0, gamma1, th, slope1, slope2, offset) = params
        a, b = tuple(
            grid - center
            for grid, center in zip(rot2d(x, y, th), rot2d(loc0, loc1, th))
        )
        return (
            I / (1 + (a / gamma0) ** 2) * 1 / (1 + (b / gamma1) ** 2)
            + x * slope1
            + y * slope2
            + offset
        )

    def set_guess(self, maximum, argmax, linparams):
        gamma0 = 5 * self.space.axes[0].res  # estimated FWHM on 10 pixels
        gamma1 = 5 * self.space.axes[1].res
        self.guess = [
            maximum,
            argmax[0],
            argmax[1],
            gamma0,
            gamma1,
            0,
            linparams[0],
            linparams[1],
            linparams[2],
        ]


class Lorentzian2Dnobkg(PeakFitBase):
    @staticmethod
    def func(grid, params):
        (x, y) = grid
        (I, loc0, loc1, gamma0, gamma1, th) = params
        a, b = tuple(
            grid - center
            for grid, center in zip(rot2d(x, y, th), rot2d(loc0, loc1, th))
        )
        return I / (1 + (a / gamma0) ** 2) * 1 / (1 + (b / gamma1) ** 2)

    def set_guess(self, maximum, argmax, linparams):
        gamma0 = 5 * self.space.axes[0].res  # estimated FWHM on 10 pixels
        gamma1 = 5 * self.space.axes[1].res
        self.guess = [maximum, argmax[0], argmax[1], gamma0, gamma1, 0]


class Lorentzian(AutoDimensionFit):
    dimensions = {1: Lorentzian1D, 2: PolarLorentzian2D}


class Gaussian1D(PeakFitBase):
    @staticmethod
    def func(grid, params):
        (x,) = grid
        (loc, I, sigma, offset, slope) = params
        return I * numpy.exp(-(((x - loc) / sigma) ** 2) / 2) + offset + x * slope


class Voigt1D(PeakFitBase):
    @staticmethod
    def func(grid, params):
        (x,) = grid
        (I, loc, sigma, gamma, slope, offset) = params
        z = (x - loc + numpy.complex(0, gamma)) / (sigma * numpy.sqrt(2))
        return (
            I * numpy.real(scipy.special.wofz(z)) / (sigma * numpy.sqrt(2 * numpy.pi))
            + offset
            + x * slope
        )

    def set_guess(self, maximum, argmax, linparams):
        gamma0 = 5 * self.space.axes[0].res  # estimated FWHM on 10 pixels
        self.guess = [maximum, argmax[0], 0.01, gamma0, linparams[0], linparams[1]]



import os

from setuptools import setup, find_packages

description = (
    "Data reduction and analysis software for two-dimensional "
    "detectors in surface X-ray diffraction"
)

long_description = """ BINoculars is a tool for data reduction and analysis of large sets of
 surface diffraction data that have been acquired with a
 two-dimensional X-ray detector. The intensity of each pixel of a
 two-dimensional detector is projected onto a three-dimensional grid
 in reciprocal-lattice coordinates using a binning algorithm. This
 allows for fast acquisition and processing of high-resolution data
 sets and results in a significant reduction of the size of the data
 set. The subsequent analysis then proceeds in reciprocal space. It
 has evolved from the specific needs of the ID03 beamline at the ESRF,
 but it has a modular design and can be easily adjusted and extended
 to work with data from other beamlines or from other measurement
 techniques."""

scripts = [
    os.path.join("scripts", d)
    for d in [
        "binoculars-fitaid",
        "binoculars-gui",
        "binoculars-processgui",
        "binoculars",
    ]
]

install_requires = ["h5py", "numpy", "matplotlib", "pyFAI", "PyGObject", "PyMca5", "PyQt5", "xrayutilities"]
# for now there is not egg info for vtk
#                    'vtk'

setup(
    name="binoculars",
    version="0.0.6-dev",
    description=description,
    long_description=long_description,
    packages=find_packages(exclude=["*.test", "*.test.*", "test.*", "test", "grafit"]),
    install_requires=install_requires,
    scripts=scripts,
    author="Willem Onderwaater, Sander Roobol, Frédéric-Emmanuel Picca",
    author_email="onderwaa@esrf.fr, picca@synchrotron-soleil.fr",
    url="FIXME",
    license="GPL-3",
    classifiers=[
        "Topic :: Scientific/Engineering",
        "Development Status :: 3 - Alpha",
        "Operating System :: POSIX",
        "Operating System :: Unix",
        "Programming Language :: Python :: 3.7",
    ],
)



import os
import sys
import argparse

from . import space, backend, util, errors


def parse_args(args):
    parser = argparse.ArgumentParser(prog="binoculars process")
    parser.add_argument(
        "-c",
        metavar="SECTION:OPTION=VALUE",
        action="append",
        type=parse_commandline_config_option,
        default=[],
        help="additional configuration option in the form section:option=value",
    )  # noqa
    parser.add_argument("configfile", help="configuration file")
    parser.add_argument("command", nargs="*", default=[])
    return parser.parse_args(args)


def parse_commandline_config_option(s):
    try:
        key, value = s.split("=", 1)
        section, option = key.split(":")
    except ValueError:
        raise argparse.ArgumentTypeError(
            "configuration specification '{0}' not in the form section:option=value".format(
                s
            )
        )  # noqa
    return section, option, value


def multiprocessing_main(xxx_todo_changeme):
    """note the double parenthesis for map() convenience"""
    (config, command) = xxx_todo_changeme
    Main.from_object(config, command)
    return config.dispatcher.destination.retrieve()


class Main(object):
    def __init__(self, config, command):
        if isinstance(config, util.ConfigSectionGroup):
            self.config = config.configfile.copy()
        elif isinstance(config, util.ConfigFile):
            self.config = config.copy()
        else:
            raise ValueError("Configfile is the wrong type")

        # distribute the configfile to space and to the metadata
        # instance
        spaceconf = self.config.copy()

        # input from either the configfile or the configsectiongroup
        # is valid
        self.dispatcher = backend.get_dispatcher(
            config.dispatcher, self, default="local"
        )
        self.projection = backend.get_projection(config.projection)
        self.input = backend.get_input(config.input)

        self.dispatcher.config.destination.set_final_options(
            self.input.get_destination_options(command)
        )  # noqa
        if "limits" in self.config.projection:
            self.dispatcher.config.destination.set_limits(
                self.config.projection["limits"]
            )  # noqa
        if command:
            self.dispatcher.config.destination.set_config(spaceconf)
        self.run(command)

    @classmethod
    def from_args(cls, args):
        args = parse_args(args)
        if not os.path.exists(args.configfile):
            # wait up to 10 seconds if it is a zpi, it might take a
            # while for the file to appear accross the network
            if not args.configfile.endswith(".zpi") or not util.wait_for_file(
                args.configfile, 10
            ):  # noqa
                raise errors.FileError(
                    "configuration file '{0}' does not exist".format(  # noqa
                        args.configfile
                    )
                )
        configobj = False
        with open(args.configfile, "rb") as fp:
            if fp.read(2) == "\x1f\x8b":  # gzip marker
                fp.seek(0)
                configobj = util.zpi_load(fp)
        if not configobj:
            # reopen args.configfile as text
            configobj = util.ConfigFile.fromtxtfile(
                args.configfile, command=args.command, overrides=args.c
            )
        return cls(configobj, args.command)

    @classmethod
    def from_object(cls, config, command):
        config.command = command
        return cls(config, command)

    def run(self, command):
        if self.dispatcher.has_specific_task():
            self.dispatcher.run_specific_task(command)
        else:
            jobs = self.input.generate_jobs(command)
            tokens = self.dispatcher.process_jobs(jobs)
            self.result = self.dispatcher.sum(tokens)
            if self.result is True:
                pass
            elif isinstance(self.result, space.EmptySpace):
                sys.stderr.write("error: output is an empty dataset\n")
            else:
                self.dispatcher.config.destination.store(self.result)

    def process_job(self, job):
        def generator():
            res = self.projection.config.resolution
            labels = self.projection.get_axis_labels()
            for intensity, weights, params in self.input.process_job(job):
                coords = self.projection.project(*params)
                if self.projection.config.limits is None:
                    yield space.Multiverse(
                        (
                            space.Space.from_image(
                                res, labels, coords, intensity, weights=weights
                            ),
                        )
                    )  # noqa
                else:
                    yield space.Multiverse(
                        space.Space.from_image(
                            res,
                            labels,
                            coords,
                            intensity,
                            weights=weights,
                            limits=limits,
                        )
                        for limits in self.projection.config.limits
                    )  # noqa

        jobverse = space.chunked_sum(generator(), chunksize=25)
        for sp in jobverse.spaces:
            if isinstance(sp, space.Space):
                sp.metadata.add_dataset(self.input.metadata)
        return jobverse

    def clone_config(self):
        config = util.ConfigSectionGroup()
        config.configfile = self.config
        config.dispatcher = self.dispatcher.config.copy()
        config.projection = self.projection.config.copy()
        config.input = self.input.config.copy()
        return config

    def get_reentrant(self):
        return multiprocessing_main


class Split(Main):
    # completely ignores the dispatcher, just yields a space per image
    def __init__(self, config, command):
        self.command = command
        if isinstance(config, util.ConfigSectionGroup):
            self.config = config.configfile.copy()
        elif isinstance(config, util.ConfigFile):
            self.config = config.copy()
        else:
            raise ValueError("Configfile is the wrong type")

        # input from either the configfile or the configsectiongroup is valid
        self.projection = backend.get_projection(config.projection)
        self.input = backend.get_input(config.input)

    def process_job(self, job):
        res = self.projection.config.resolution
        labels = self.projection.get_axis_labels()
        for intensity, weights, params in self.input.process_job(job):
            coords = self.projection.project(*params)
            if self.projection.config.limits is None:
                yield space.Space.from_image(
                    res, labels, coords, intensity, weights=weights
                )  # noqa
            else:
                yield space.Multiverse(
                    space.Space.from_image(
                        res, labels, coords, intensity, weights=weights, limits=limits
                    )
                    for limits in self.projection.config.limits
                )  # noqa

    def run(self):
        for job in self.input.generate_jobs(self.command):
            for verse in self.process_job(job):
                yield verse



# -*- coding: utf-8 -*-
"""
Created on Fri Dec 06 12:15:46 2013

@author: prevot

Objectif: remplacer les courbes par des instances de classe


FitCurve.nom : nom de la courbe
FitCurve.function : fonction associee
FitCurve.parameters : parametres standards de la courbe
FitCurve.letter : lettre symbolique de la fonction
FitCurve.getIntegral() : calcule l'integrale pour les fonctions le permettant
FitCurve.RectangularActionTool : le type d'outil de tracage
FitCurve.action : le comportement de la souris au traçage

"""

import copy
import numpy as np

from os import path, getcwd

from guiqwt.plot import CurveWidget, PlotManager
from guiqwt.builder import make

from guidata.qt.QtCore import Qt, SIGNAL, QRegExp, QRect, QSize, QPoint, QString
from guidata.qt.QtGui import (
    QSyntaxHighlighter,
    QTextCharFormat,
    QListWidget,
    QListWidgetItem,
    QTextEdit,
    QTableWidget,
    QTableWidgetItem,
    QLabel,
    QMenu,
    QAction,
    QCursor,
    QColor,
    QIcon,
    QMessageBox,
    QInputDialog,
    QFileDialog,
    QDialog,
    QMainWindow,
    QSplitter,
    QApplication,
    QPushButton,
    QButtonGroup,
    QSpinBox,
    QRadioButton,
    QCheckBox,
)
from guidata.qthelpers import add_actions, get_std_icon

import sys, weakref
from guiqwt.events import RectangularSelectionHandler, setup_standard_tool_filter
from guiqwt.tools import (
    OpenFileTool,
    CommandTool,
    DefaultToolbarID,
    RectangularActionTool,
    BaseCursorTool,
)
from guiqwt.shapes import XRangeSelection
from guiqwt.signals import SIG_END_RECT, SIG_RANGE_CHANGED, SIG_ITEM_REMOVED
from guiqwt.config import _
from guiqwt.interfaces import ICurveItemType, IShapeItemType
from lmfit import minimize, Parameters, Parameter
from scipy.special import erf
from scipy.integrate import simps

SHAPE_Z_OFFSET = 1000

try:
    _fromUtf8 = QString.fromUtf8
except AttributeError:
    _fromUtf8 = lambda s: s

abspath = path.abspath(__file__)
dirpath = path.dirname(abspath)

abspath = getcwd()
runpath = path.join(dirpath, "Run.png")
gausspath = path.join(dirpath, "Gauss.png")
eraserpath = path.join(dirpath, "Eraser.png")
steppath = path.join(dirpath, "Step.png")
doorpath = path.join(dirpath, "Door.png")
gaussderpath = path.join(dirpath, "GaussDer.png")
gausslorpath = path.join(dirpath, "Skew.png")
skewpath = path.join(dirpath, "Skew.png")
yfullrangepath = path.join(dirpath, "y_full_range.png")


class Highlighter(QSyntaxHighlighter):
    PATTERNS = [" ", "\t", ","]

    def __init__(self, parent, patternid=0):
        super(Highlighter, self).__init__(parent)
        self.highlightFormat = QTextCharFormat()
        self.highlightFormat.setForeground(Qt.blue)
        self.highlightFormat.setBackground(Qt.green)
        self.setpattern(patternid)

    def setpattern(self, patternid):
        self.expression = QRegExp(self.PATTERNS[patternid])

    def highlightBlock(self, text):
        index = self.expression.indexIn(text)
        while index >= 0:
            length = self.expression.matchedLength()
            self.setFormat(index, length, self.highlightFormat)
            index = text.indexOf(self.expression, index + length)


# fonctions de fit
def lorentz(x, p):
    # p[0] max d'intensite, p[1] centre, p[2],FWHM
    return p[0] / (1.0 + ((x - p[1]) / 0.5 / p[2]) ** 2)


def lorentzder(x, p):
    # p[0] max d'intensite, p[1] centre, p[2],FWHM
    return p[0] * (p[1] - x) / p[2] / (1.0 + ((x - p[1]) / 0.5 / p[2]) ** 2) ** 2


def gauss(x, p):
    w = 0.600561204393225
    return p[0] * np.exp(-(((x - p[1]) / w / p[2]) ** 2))


def gausslor(x, p):
    # gaussian switching to lorentzian for x-p[1]=p[2]*p[3]
    if p[3] > 0:
        w = 0.600561204393225
        p32 = p[3] * p[2]
        pp1 = p32 + p[1] - ((w * p[2]) ** 2) / p32
        pp0 = p[0] * np.exp(-(((p32) / w / p[2]) ** 2)) * (p[1] + p32 - pp1) ** 2
        b = np.greater(x, p[1] + p32)
        return (p[0] * np.exp(-(((x - p[1]) / w / p[2]) ** 2))) * (1 - b) + (
            pp0 / (x - pp1) ** 2
        ) * b
    elif p[3] < 0:
        w = 0.600561204393225
        p32 = p[3] * p[2]
        pp1 = p32 + p[1] - ((w * p[2]) ** 2) / p32
        pp0 = p[0] * np.exp(-(((p32) / w / p[2]) ** 2)) * (p[1] + p32 - pp1) ** 2
        b = np.less(x, p[1] + p32)
        return (p[0] * np.exp(-(((x - p[1]) / w / p[2]) ** 2))) * (1 - b) + (
            pp0 / (x - pp1) ** 2
        ) * b
    else:
        return gauss(x, p)


def gaussder(x, p):
    # derivee d'une gaussienne
    w = 0.600561204393225
    return p[0] * (p[1] - x) / p[2] * np.exp(-(((x - p[1]) / w / p[2]) ** 2))


def skew(x, p):
    w = 0.600561204393225
    return (
        p[0]
        * np.exp(-(((x - p[1]) / w / p[2]) ** 2))
        * (1.0 + erf((x - p[1]) * p[3] / w / p[2]))
    )


def voigt(x, p):
    w = 0.600561204393225
    return p[0] * (
        (1.0 - p[3]) * np.exp(-(((x - p[1]) / w / p[2]) ** 2))
        + p[3] / (1.0 + ((x - p[1]) / 0.5 / p[2]) ** 2)
    )


def step(x, p):
    w = 0.600561204393225
    return 0.5 * p[0] * (1.0 - erf((x - p[1]) / w / p[2]))


def door(x, p):
    return (
        p[0]
        / 2.0
        * (erf((x - p[1] + p[2] / 2.0) / p[3]) - erf((x - p[1] - p[2] / 2.0) / p[3]))
    )


# integrales des fonctions
# integrales des fonctions
def Ilorentz(p):
    return p[0] * p[2] * np.pi / 2.0


def Igauss(p):
    wpi = 1.064467019431226
    return p[0] * p[2] * wpi


def Ivoigt(p):
    wpi = 1.064467019431226
    return p[0] * p[2] * ((1.0 - p[3]) * wpi + p[3] * np.pi / 2.0)


def Igausslor(p):
    w = 0.600561204393225
    wpis2 = 0.532233509715613

    p32 = np.abs(p[3] * p[2])
    pp = ((w * p[2]) ** 2) / p32
    pp1 = p32 + p[1] - pp

    pp0 = p[0] * np.exp(-(((p32) / w / p[2]) ** 2)) * (p[1] + p32 - pp1) ** 2
    return p[0] * p[2] * wpis2 * (1.0 + erf(p[3] / p[2])) + pp0 / pp


# dictionnaire reliant la fonction au nom
curvetypes = dict(
    gaussian=gauss,
    lorentzian=lorentz,
    voigt=voigt,
    door=door,
    step=step,
    gaussian_derivative=gaussder,
    lorentz_derivative=gaussder,
    gaussian_lorentzian=gausslor,
    skew=skew,
)
# dictionnaire reliant l'integrale au nom
inttypes = dict(
    gaussian=Igauss,
    lorentzian=Ilorentz,
    voigt=Ivoigt,
    door=None,
    step=None,
    gaussian_derivative=None,
    lorentz_derivative=None,
    gaussian_lorentzian=Igausslor,
    skew=Igauss,
)

curvenames = list(
    (
        "gaussian",
        "lorentzian",
        "voigt",
        "door",
        "step",
        "gaussian_derivative",
        "lorentz_derivative",
        "gaussian_lorentzian",
        "skew",
    )
)
curveparams = dict(
    gaussian=(
        ("int", 1.0, None, None),
        ("pos", 0.0, None, None),
        ("fwhm", 1.0, 0.0, None),
    ),
    lorentzian=(
        ("int", 1.0, None, None),
        ("pos", 0.0, None, None),
        ("fwhm", 1.0, 0.0, None),
    ),
    voigt=(
        ("int", 1.0, None, None),
        ("pos", 0.0, None, None),
        ("fwhm", 1.0, 0, None),
        ("eta", 0.5, 0.0, 1.0),
    ),
    door=(
        ("int", 1.0, None, None),
        ("pos", 0.0, None, None),
        ("fwhm", 1.0, 0, None),
        ("steep", 0.25, 0.0, None),
    ),
    step=(
        ("int", 1.0, None, None),
        ("pos", 0.0, None, None),
        ("steep", 1.0, 0.0, None),
    ),
    gaussian_derivative=(
        ("int", 1.0, None, None),
        ("pos", 0.0, None, None),
        ("fwhm", 1.0, 0.0, None),
    ),
    lorentzian_derivative=(
        ("int", 1.0, None, None),
        ("pos", 0.0, None, None),
        ("fwhm", 1.0, 0.0, None),
    ),
    gaussian_lorentzian=(
        ("int", 1.0, None, None),
        ("pos", 0.0, None, None),
        ("fwhm", 1.0, 0.0, None),
        ("eps", 0.1, None, None),
    ),
    skew=(
        ("int", 1.0, None, None),
        ("pos", 0.0, None, None),
        ("fwhm", 1.0, 0.0, None),
        ("alpha", 0.0, None, None),
    ),
)

# parametres de reference communs aux differentes courbes
int0 = Parameter("int", 1.0, min=0.0)
pos0 = Parameter("pos", 0.0)
fwhm0 = Parameter("fwhm", 1.0, min=0.0)
eta0 = Parameter("eta", 0.5, min=0.0, max=1.0)
steep0 = Parameter("steep", 0.25, min=0.0)
alpha0 = Parameter("alpha", 0.0)
eps0 = Parameter("alpha", 0.0)


class FitCurve:
    # definit une classe pour les fonction de fit
    def __init__(self, name, curvetype, list_of_parameters, integral=None):
        self.name = name
        self.curvetype = curvetype
        self.Parameters = Parameters()
        for parameter in list_of_parameters:
            self.Parameters.__setitem__(parameter.name, parameter)


gaussparams = FitCurve("gaussian", gauss, (int0, pos0, fwhm0), Igauss)
lorentzparams = FitCurve("lorentzian", lorentz, (int0, pos0, fwhm0), Ilorentz)
voigtparams = FitCurve("voigt", voigt, (int0, pos0, fwhm0, eta0), Ivoigt)
doorparams = FitCurve("door", door, (int0, pos0, fwhm0, steep0))
stepparams = FitCurve("step", step, (int0, pos0, steep0))
gaussderparams = FitCurve("gaussian_derivative", gaussder, (int0, pos0, fwhm0))
lorentzderparams = FitCurve("lorentzian_derivative", gaussder, (int0, pos0, fwhm0))
gausslorparams = FitCurve("gaussian_lorentzian", gausslor, (int0, pos0, fwhm0, eps0))
skewparams = FitCurve("skew", skew, (int0, pos0, fwhm0, alpha0))


def model(params, x, fixparams):
    # fixparams : parametres fixes
    # nombre de courbes utilisees
    degpol = fixparams[0]
    curveslist = fixparams[1]

    ncurv = len(curveslist)
    # degre du polynome : degpol

    p = np.empty(degpol + 1)
    for i in range(degpol + 1):
        name = "pol%d" % (i)
        p[i] = params[name].value

    yth = poly(x, p, degpol)

    for i in range(ncurv):
        i1 = i + 1
        kstr = "%d" % (i1)
        curvename = curveslist[i]
        p = list()
        for cp in curveparams[curvename]:
            entry = cp[0] + kstr
            p.append(params[entry].value)
        yth = yth + curvetypes[curvename](x, p)
    return yth


def integral(params, curveslist):
    # retourne la liste des valeurs des integrales
    # le premier point est l'integrale totale

    ncurv = len(curveslist)
    intlist = list()
    inttot = 0.0
    for i in range(ncurv):
        i1 = i + 1
        kstr = "%d" % (i1)
        curvename = curveslist[i]
        p = list()
        for cp in curveparams[curvename]:
            ent = cp[0]
            entry = ent + kstr
            p.append(params[entry].value)
        intfunct = inttypes[curvename]
        if intfunct is not None:
            intpart = intfunct(p)  # on recupere la fonction qui calcule l'integrale
            intlist.append(intpart)
            inttot = inttot + intpart
        else:
            intlist.append(None)
    intlist.insert(0, inttot)
    return intlist


def residual(params, x, data, fixparams):
    # fparam : parametres fixes
    return model(params, x, fixparams) - data


## Parametric function: 'p' is the parameter vector, 'x' the independent varibles
"""
def poly(x,p,degpol):
    y=p[0]
    xp=1
    for i in range(degpol):
        xp=xp*x
        y=y+xp*p[i+1]
    return y
"""


def poly(x, p, degpol):
    y = 0
    # p[0]
    xp = 1
    for i in range(degpol + 1):
        y = y + xp * p[i]
        xp = xp * x
    return y


class TagsWindow(QDialog):
    # definit une fenetre pour rentrer l'affichage des tags
    def __init__(self, tags):

        QDialog.__init__(self)
        self.setWindowTitle(_fromUtf8("List of tags"))
        self.list = QListWidget(self)

        self.list = QListWidget(self)
        self.list.setGeometry(QRect(5, 5, 200, 300))
        for tag in tags:

            tagname = tag[0]
            tagvalue = ": %f" % tag[1]
            text = tagname + tagvalue
            item = QListWidgetItem(self.list)
            item.setText(text)

        self.OK = QPushButton(self)
        self.OK.setGeometry(QRect(5, 310, 100, 25))
        self.OK.setText("OK")

        self.setGeometry(QRect(5, 30, 210, 350))

        self.connect(self.OK, SIGNAL(_fromUtf8("clicked()")), self.appl)

        self.exec_()

    def appl(self):
        self.close()


class readfileparams:
    # definit une classe pour la lecture des parametres d'un fichier a lire
    def __init__(self):
        self.heading = 0  # taille de l'entete
        self.title = False  # ligne de titre
        self.delimiter = 0  # separateur de nombre
        self.x = 1  # indice colonne x
        self.y = 2  # indice colonne y


class readfilewindow(QDialog):

    SIG_FIT_DONE = Signal("fit_done")

    def __init__(self, params, ficlines):
        self.params = params
        super(
            readfilewindow, self
        ).__init__()  # permet l'initialisation de la fenetre sans perdre les fonctions associees
        self.setWindowTitle("Parameters for reading file")
        self.setFixedSize(QSize(410, 320))
        self.ficlines = ficlines

        self.display = QTextEdit(self)
        self.display.setReadOnly(True)
        self.display.setGeometry(QRect(5, 5, 400, 100))
        self.display.setLineWrapMode(0)
        self.highlighter = Highlighter(self.display.document(), self.params.delimiter)

        self.lab1 = QLabel(self)
        self.lab1.setGeometry(QRect(5, 110, 50, 25))
        self.lab1.setText("heading")

        self.lab2 = QLabel(self)
        self.lab2.setGeometry(QRect(5, 140, 50, 25))
        self.lab2.setText("title")

        self.lab3 = QLabel(self)
        self.lab3.setGeometry(QRect(5, 170, 50, 25))
        self.lab3.setText("delimiter")

        self.lab4 = QLabel(self)
        self.lab4.setGeometry(QRect(5, 200, 50, 25))
        self.lab4.setText("x column")

        self.lab4 = QLabel(self)
        self.lab4.setGeometry(QRect(5, 230, 50, 25))
        self.lab4.setText("y column")

        self.heading = QSpinBox(self)
        self.heading.setGeometry(QRect(60, 110, 45, 25))
        self.heading.setValue(params.heading)
        self.heading.setMaximum(len(ficlines))

        self.title = QRadioButton(self)
        self.title.setGeometry(QRect(60, 140, 45, 25))
        self.title.setChecked(params.title)

        self.delimiter0 = QCheckBox(self)
        self.delimiter0.setGeometry(QRect(60, 170, 65, 25))
        self.delimiter0.setText("space")

        self.delimiter1 = QCheckBox(self)
        self.delimiter1.setGeometry(QRect(130, 170, 65, 25))
        self.delimiter1.setText("tab")

        self.delimiter2 = QCheckBox(self)
        self.delimiter2.setGeometry(QRect(200, 170, 45, 25))
        self.delimiter2.setText("comma")

        self.delimiters = QButtonGroup(self)

        self.delimiters.addButton(self.delimiter0, 0)
        self.delimiters.addButton(self.delimiter1, 1)
        self.delimiters.addButton(self.delimiter2, 2)
        self.delimiters.button(params.delimiter).setChecked(True)

        self.x = QSpinBox(self)
        self.x.setGeometry(QRect(60, 200, 45, 25))
        self.x.setValue(params.x)

        self.y = QSpinBox(self)
        self.y.setGeometry(QRect(60, 230, 45, 25))
        self.y.setValue(params.y)

        self.OK = QPushButton(self)
        self.OK.setGeometry(QRect(5, 260, 90, 25))
        self.OK.setText("OK")

        self.Cancel = QPushButton(self)
        self.Cancel.setGeometry(QRect(100, 260, 90, 25))
        self.Cancel.setText("Cancel")

        self.connect(
            self.heading, SIGNAL(_fromUtf8("valueChanged(int)")), self.changeheading
        )
        self.QObject.connect(self.Cancel, SIGNAL(_fromUtf8("clicked()")), self.closewin)
        self.QObject.connect(self.OK, SIGNAL(_fromUtf8("clicked()")), self.appl)
        self.QObject.connect(
            self.delimiters, SIGNAL(_fromUtf8("buttonClicked(int)")), self.setpattern
        )
        self.changeheading(self.params.heading)

        self.exec_()

    def setpattern(self, patternid):
        self.highlighter.setpattern(patternid)
        self.display.setText(self.text)

    def changeheading(self, value):
        self.params.heading = value
        # first line
        n1 = min(value, len(self.ficlines) - 1)
        n2 = min(value + 20, len(self.ficlines))
        self.text = ""
        for i in range(n1, n2):
            self.text = self.text + self.ficlines[i]

        self.display.setText(self.text)

    def appl(self):
        # try:
        self.params.heading = int(self.heading.text())
        self.params.title = self.title.isChecked()
        self.params.x = int(self.x.text())
        self.params.y = int(self.y.text())
        self.params.delimiter = self.delimiters.checkedId()

        # except Exception:
        #    QMessageBox.about(self, 'Error','Input can only be a number')
        #    return

        self.close()

    def closewin(self):
        self.close()


class FitTable(QTableWidget):
    # objet qui gere les entree de parametres, l'affichage des courbes, le lancement des fits, l'enregistrement du resultat
    def __init__(self, parent=None, extendedparams=None):
        QTableWidget.__init__(self, parent=parent)

        self.setColumnCount(8)
        self.setObjectName(_fromUtf8("tableValeurs"))
        self.setHorizontalHeaderLabels(
            (
                "Parameters",
                "Estimation",
                "Fit result",
                "Sigma",
                "Restrains",
                "Min",
                "Max",
                "Expression",
            )
        )

        # on remet le tableau a zero : deux rangees pour un polynome du 1er degre
        self.reset(extendedparams)

        # on definit les differents menus contextuels
        self.initcontexts()

        # au cas ou
        self.int1 = 0

        # utilise pour contourner le signal de cellChanged
        self.trackchange = True

        # On connecte les signaux
        self.connect(self, SIGNAL(_fromUtf8("cellChanged(int,int)")), self.cellChanged)
        self.connect(self, SIGNAL(_fromUtf8("cellPressed(int,int)")), self.cellPressed)
        self.connect(
            self.horizontalHeader(),
            SIGNAL(_fromUtf8("sectionPressed(int)")),
            self.sectionClicked,
        )

        # pas de fit sauve
        self.isfitted = False
        self.issaved = False
        self.tags = list()
        self.saveint = False
        self.ints = []
        self.updatestart = (
            False  # si coche, on active setwholeastry apres enregistrement
        )
        self.cwd = abspath

    def reset(self, extendedparams=None):
        if extendedparams is None:
            # on met deux items vide
            extendedparams = ExtendedParams()
            extendedparams.addbackground()
            extendedparams.addbackground()
            # on remplit le tableau a l'aide du dictionnaire extendedparams
        self.initparams(extendedparams, settry=True)
        self.savename = None
        self.isfitted = False

    def inititems(self, listrow):
        # met pour les lignes i de listrow un item vide non editable pour colonnes 0,2,3,4
        for i in listrow:
            for j in range(8):
                # self.setItem(i,j,QTableWidgetItem(_fromUtf8("")))
                self.setItem(i, j, QTableWidgetItem(0))
            self.item(i, 0).setFlags(Qt.ItemFlags(33))
            self.item(i, 2).setFlags(Qt.ItemFlags(33))
            self.item(i, 3).setFlags(Qt.ItemFlags(33))
            self.item(i, 4).setFlags(Qt.ItemFlags(33))

    def initparams(self, extendedparams=None, settry=False, setopt=False):
        self.trackchange = False

        # on redessine le tableau a l'aide du dictionnaire extendedparams
        # si settry, on remplit aussi les valeurs d'essai, les bornes, les expressions
        if extendedparams is not None:
            # on utilise les valeurs donnees en input, sinon on prend celles de self.extendedparams
            self.extendedparams = extendedparams

        n = 0
        oldrc = self.rowCount()
        newrc = len(self.extendedparams.partry) - len(
            self.extendedparams.tags
        )  # on n'affiche pas les tags
        self.setRowCount(newrc)

        if newrc > oldrc:
            # on rajoute des rangees vides
            self.inititems(range(oldrc, newrc))

        for k in range(self.extendedparams.degpol + 1):
            kstr = "%d" % (k)
            entry = "pol" + kstr
            self.setrow(entry, n, settry, setopt)
            n = n + 1

        for k in range(self.extendedparams.ncurves):
            curvename = self.extendedparams.curveslist[k]
            for cp in curveparams[curvename]:
                ent = cp[0]
                entry = ent + "%d" % (k + 1)
                self.setrow(entry, n, settry, setopt)
                n = n + 1
        self.trackchange = True

    def setrow(self, nom, n, settry=True, setopt=False):
        # appele uniquement dans initparams. donc on ne change pas self.trackchange
        # on ne change pas expr car on veut pouvoir la garder en memoire meme si on ne l'utilise pas
        ptry = self.extendedparams.partry
        popt = self.extendedparams.paropt
        self.item(n, 0).setText(_fromUtf8(nom))
        """
        self.setItem(n,0,QTableWidgetItem(_fromUtf8(nom)))
        self.setItem(n,2,QTableWidgetItem(_fromUtf8("")))
        self.setItem(n,3,QTableWidgetItem(_fromUtf8("")))
        self.setItem(n,4,QTableWidgetItem(_fromUtf8("")))
        self.item(n,0).setFlags(Qt.ItemFlags(33))
        self.item(n,2).setFlags(Qt.ItemFlags(33))
        self.item(n,3).setFlags(Qt.ItemFlags(33))
        self.item(n,4).setFlags(Qt.ItemFlags(33))
        """

        if settry:
            vtry = ptry[nom].value

            mintry = ptry[nom].min
            maxtry = ptry[nom].max
            print(nom, mintry, maxtry)
            varytry = ptry[nom].vary
            exprtry = ptry[nom].expr
            self.item(n, 1).setText(_fromUtf8("%f" % (vtry)))
            # self.setItem(n,1,QTableWidgetItem(_fromUtf8("%f" % (vtry))))

            if exprtry is not None:
                # self.setItem(n,4,QTableWidgetItem(_fromUtf8(exprtry)))
                self.item(n, 4).setText(_fromUtf8("Expr"))
            elif varytry is True:
                # self.setItem(n,4,QTableWidgetItem(_fromUtf8("Free")))
                self.item(n, 4).setText(_fromUtf8("Free"))
            else:
                # self.setItem(n,4,QTableWidgetItem(_fromUtf8("Fixed")))
                self.item(n, 4).setText(_fromUtf8("Fixed"))

            if mintry is not None:
                # self.setItem(n,5,QTableWidgetItem(_fromUtf8("%f" % (mintry))))
                self.item(n, 5).setText(_fromUtf8("%f" % (mintry)))

            if maxtry is not None:
                # self.setItem(n,6,QTableWidgetItem(_fromUtf8("%f" % (maxtry))))
                self.item(n, 6).setText(_fromUtf8("%f" % (maxtry)))

        # here we rewrite the expression even if it is not used
        self.item(n, 7).setText(_fromUtf8(ptry[nom].express))

        if setopt and self.isfitted:
            vopt = popt[nom].value
            # self.setItem(n,2,QTableWidgetItem(_fromUtf8("%f" % (vopt))))
            self.item(n, 2).setText(_fromUtf8("%f" % (vopt)))
            verr = popt[nom].stderr
            if verr is not None:
                self.item(n, 3).setText(_fromUtf8("%f" % (verr)))

        else:
            self.item(n, 2).setText(_fromUtf8(""))
            self.item(n, 3).setText(_fromUtf8(""))
        n = n + 1

        """
        exemple:
        self["degpol"]=1
        self["ncurves"]=2
        self["curveslist"]=({courbe:gaussienne,parametre:(int,pos,fwhm)},{courbe:voigt,parametre:(int,pos,fwhm,eta)})
        self["partry"]=orderer dict
        """

    def setactivecurve(self, val):
        # defini la courbe active qui va etre modifie a la souris
        self.activecurve = val

    def initcontexts(self):
        # initialise une liste de menus contextuels pour les cases, associes a chaque colonne
        self.contexts = list(None for i in range(self.columnCount()))
        # initialise une liste de menus contextuels pour les sectionheaders, associes a chaque colonne
        self.Contexts = list(None for i in range(self.columnCount()))

        # menu pour la colonne 0
        contexta = QMenu(self)
        self.addbg = QAction("Increase background order", self)
        contexta.addAction(self.addbg)
        self.connect(self.addbg, SIGNAL(_fromUtf8("triggered()")), self.add_bg)

        self.rmbg = QAction("Decrease background order", self)
        contexta.addAction(self.rmbg)
        self.connect(self.rmbg, SIGNAL(_fromUtf8("triggered()")), self.rm_bg)

        self.rmbg = QAction("Decrease background order", self)
        contexta.addAction(self.rmbg)
        self.connect(self.rmbg, SIGNAL(_fromUtf8("triggered()")), self.rm_bg)

        self.freezebg = QAction("Fixed background", self)
        contexta.addAction(self.freezebg)
        self.connect(self.freezebg, SIGNAL(_fromUtf8("triggered()")), self.freeze_bg)

        self.releasebg = QAction("Free background", self)
        contexta.addAction(self.releasebg)
        self.connect(self.releasebg, SIGNAL(_fromUtf8("triggered()")), self.release_bg)

        contextb = QMenu(self)
        self.addcv = QAction("Add a curve", self)
        contextb.addAction(self.addcv)
        self.connect(self.addcv, SIGNAL(_fromUtf8("triggered()")), self.set_cv)

        self.rmcv = QAction("Remove this curve", self)
        contextb.addAction(self.rmcv)
        self.connect(self.rmcv, SIGNAL(_fromUtf8("triggered()")), self.rm_cv)

        self.chcv = QAction("Reconfigure curve", self)
        contextb.addAction(self.chcv)
        self.connect(self.chcv, SIGNAL(_fromUtf8("triggered()")), self.ch_cv)

        self.freezecv = QAction("Fixed curve", self)
        contextb.addAction(self.freezecv)
        self.connect(self.freezecv, SIGNAL(_fromUtf8("triggered()")), self.freeze_cv)

        self.releasecv = QAction("Free curve", self)
        contextb.addAction(self.releasecv)
        self.connect(self.releasecv, SIGNAL(_fromUtf8("triggered()")), self.release_cv)

        self.contexts[0] = (contexta, contextb)

        # menu pour la colonne 2

        context = QMenu(self)

        self.setastry = QAction("Set as try", self)
        context.addAction(self.setastry)
        self.connect(self.setastry, SIGNAL(_fromUtf8("triggered()")), self.set_as_try)

        self.showfit = QAction("Display Fit", self)
        context.addAction(self.showfit)
        self.connect(self.showfit, SIGNAL(_fromUtf8("triggered()")), self.showfitpart)

        self.showtry = QAction("Display Try", self)
        context.addAction(self.showtry)
        self.connect(self.showtry, SIGNAL(_fromUtf8("triggered()")), self.drawtrypart)

        self.contexts[2] = context

        # menu pour la colonne 4
        context = QMenu(self)

        self.setfixed = QAction("Fixed", self)
        context.addAction(self.setfixed)
        self.connect(self.setfixed, SIGNAL(_fromUtf8("triggered()")), self.set_fixed)

        self.setfree = QAction("Free", self)
        context.addAction(self.setfree)
        self.connect(self.setfree, SIGNAL(_fromUtf8("triggered()")), self.set_free)

        self.useexpr = QAction("Expr", self)
        context.addAction(self.useexpr)
        self.connect(self.useexpr, SIGNAL(_fromUtf8("triggered()")), self.use_expr)

        self.contexts[4] = context

        # menu pour le header 2
        context = QMenu(self)
        self.setwholeastry = QAction("Set whole as try", self)
        context.addAction(self.setwholeastry)
        self.connect(
            self.setwholeastry, SIGNAL(_fromUtf8("triggered()")), self.set_wholeastry
        )

        self.Contexts[2] = context

        # menu pour le header 4
        context = QMenu(self)
        self.inverserestrains = QAction("Inverse restrains", self)
        context.addAction(self.inverserestrains)
        self.connect(
            self.inverserestrains,
            SIGNAL(_fromUtf8("triggered()")),
            self.inverse_restrains,
        )

        self.fixall = QAction("All fixed", self)
        context.addAction(self.fixall)
        self.connect(self.fixall, SIGNAL(_fromUtf8("triggered()")), self.fix_all)

        self.releaseall = QAction("All free", self)
        context.addAction(self.releaseall)
        self.connect(
            self.releaseall, SIGNAL(_fromUtf8("triggered()")), self.release_all
        )

        self.Contexts[4] = context

    def cellPressed(self, int1, int2):
        self.int1 = int1

        context = self.contexts[int2]

        if context is not None:

            if int2 == 0:
                if int1 <= self.extendedparams.degpol:
                    context[0].popup(QCursor.pos())
                else:
                    context[1].popup(QCursor.pos())
            else:
                context.popup(QCursor.pos())
        # self.extendedparams.printparams()

    def cellChanged(self, int1, int2):

        if self.trackchange:
            # print  "Changed at",int1,int2
            if int2 == 1:  # on change la valeur du parametre
                txt = str(self.item(int1, 1).text())
                entry = str(self.item(int1, 0).text())
                self.extendedparams.partry[entry].value = float(txt)
                self.drawtry()

            if int2 == 5:  # on change la borne inferieure
                txt = str(self.item(int1, 5).text())
                entry = str(self.item(int1, 0).text())
                self.extendedparams.partry[entry].min = float(txt)

            if int2 == 6:  # on change la borne superieure
                txt = str(self.item(int1, 6).text())
                entry = str(self.item(int1, 0).text())
                self.extendedparams.partry[entry].max = float(txt)

            if int2 == 7:  # on change l'expression
                txt = str(self.item(int1, 7).text())
                entry = str(self.item(int1, 0).text())
                self.extendedparams.partry[entry].express = txt  # a copy of expression
                if len(txt) == 0:
                    self.extendedparams.partry[entry].expr = None
                    self.extendedparams.partry[entry].vary = True
                else:
                    self.extendedparams.partry[entry].expr = txt

                    self.trackchange = False
                    self.item(int1, 4).setText(_fromUtf8("Expr"))
                    self.trackchange = True

    def sectionClicked(self, int2):
        # quand une colonne entiere est selectionnee valable pour les colonnes 2 et 4
        context = self.Contexts[int2]
        if context is not None:
            context.popup(QCursor.pos())
        # self.extendedparams.printparams()

    def add_bg(self):
        # on ajoute un ordre au polynome de fond
        self.extendedparams.addbackground()
        self.isfitted = False
        self.initparams(settry=True)
        self.drawtry()

    def rm_bg(self):
        # on enleve un ordre au polynome de fond
        self.extendedparams.removebackground()
        self.isfitted = False
        self.initparams(settry=True)
        self.drawtry()

    def freeze_bg(self):
        # on met fixe tous les parametres du fond
        self.trackchange = False
        for int1 in range(self.extendedparams.degpol + 1):
            item = self.item(int1, 4)
            item.setText(_fromUtf8("Fixed"))
            entry = str(self.item(int1, 0).text())
            self.extendedparams.partry[entry].vary = False
        self.trackchange = True

    def release_bg(self):
        # on met fixe tous les parametres du fond
        self.trackchange = False
        for int1 in range(self.extendedparams.degpol + 1):
            item = self.item(int1, 4)
            item.setText(_fromUtf8("Free"))
            entry = str(self.item(int1, 0).text())
            self.extendedparams.partry[entry].vary = True
        self.trackchange = True

    def freeze_cv(self):
        # on met fixe tous les parametres d'une courbe
        self.trackchange = False
        entry = str(self.item(self.int1, 0).text())
        icurve = self.extendedparams.partry[entry].number
        self.extendedparams.freezecurve(icurve)
        self.initparams(settry=True, setopt=True)
        self.trackchange = True

    def release_cv(self):
        # on met fixe tous les parametres d'une courbe
        self.trackchange = False
        entry = str(self.item(self.int1, 0).text())
        icurve = self.extendedparams.partry[entry].number
        self.extendedparams.releasecurve(icurve)
        self.initparams(settry=True, setopt=True)
        self.trackchange = True

    def release_all(self):
        # on met fixe tous les parametres du fond
        self.trackchange = False
        for int1 in range(self.rowCount()):
            item = self.item(int1, 4)
            item.setText(_fromUtf8("Free"))
            entry = str(self.item(int1, 0).text())
            self.extendedparams.partry[entry].vary = True
        self.trackchange = True

    def fix_all(self):
        # on met fixe tous les parametres du fond
        self.trackchange = False
        for int1 in range(self.rowCount()):
            item = self.item(int1, 4)
            item.setText(_fromUtf8("Fixed"))
            entry = str(self.item(int1, 0).text())
            self.extendedparams.partry[entry].vary = False
        self.trackchange = True

    def set_cv(self, curvetype=None, params=None, bg=0.0):
        # icurve numero de la courbe a changer le cas echeant -1 c'est la derniere
        print("bg", bg)
        if curvetype is None:
            # on ouvre une fenetre pour demander le choix de courbe
            self.isfitted = False
            cvtyp, ok = QInputDialog.getItem(
                self, "curve type", "curve choice", curvenames, editable=False
            )
            if ok:
                self.extendedparams.addcurve(str(cvtyp))

                self.initparams(settry=True)
                self.drawtry()
        else:
            # on prend les donnees en parametre
            if self.activecurve == "new":
                self.isfitted = False
                # plot.lastcurve=make.curve(xtry, ytry, color="k",title='gaussienne')
                # plot.add_item(plot.lastcurve)
                self.extendedparams.addcurve(curvetype, params)
                # si la courbe est la premiere courbe, on met le fond a la valeur bg
                icurve = len(self.extendedparams.curveslist) - 1
                if icurve == 0:
                    self.extendedparams.partry["pol0"].value = bg
                self.initparams(settry=True)
                self.plot.newcurve = False
                self.activecurve = -1
            else:
                # plot.lastcurve.set_data(xtry, ytry)
                self.extendedparams.changecurve(
                    paramvalues=params, icurve=self.activecurve
                )
                print(self.activecurve)
                # si la courbe est la premiere courbe, on met le fond a la valeur bg
                if self.activecurve == 0:
                    self.extendedparams.partry["pol0"].value = bg
                elif self.activecurve == -1:
                    icurve = len(self.extendedparams.curveslist) - 1
                    if icurve == 0:
                        self.extendedparams.partry["pol0"].value = bg
                self.initparams(settry=True)
            self.drawtry()

    def rm_cv(self):
        # on supprime la courbe selectionnee
        entry = str(self.item(self.int1, 0).text())
        icurve = self.extendedparams.partry[entry].number
        self.isfitted = False
        if icurve > 0:  # normalement c'est toujours le cas
            self.extendedparams.removecurve(icurve)
            self.initparams(settry=True)
            self.drawtry()
        self.parent().parent().manager.activate_default_tool()

    def ch_cv(self):
        # a partir de la souris, on retrace la courbe selectionnee
        entry = str(self.item(self.int1, 0).text())
        i = self.extendedparams.partry[entry].number
        if i > 0:  # normalement c'est toujours le cas
            curvename = self.extendedparams.curveslist[i - 1]
            manager = self.parent().parent().manager
            self.activecurve = i - 1
            fittool = manager.get_tool(FitTool)
            if curvename == "gaussian":
                fittool.gausstool.activate()
            elif curvename == "lorentzian":
                fittool.lorentztool.activate()
            elif curvename == "voigt":
                fittool.voigttool.activate()
            elif curvename == "step":
                fittool.steptool.activate()
            elif curvename == "door":
                fittool.doortool.activate()
            elif curvename == "gaussian_derivative":
                fittool.gaussdertool.activate()
            elif curvename == "lorentzian_derivative":
                fittool.lorentzdertool.activate()
            elif curvename == "gaussian_lorentzian":
                fittool.gausslortool.activate()
            elif curvename == "skew":
                fittool.skewtool.activate()

            self.initparams(settry=True)
            self.drawtry()

        # self.gausscurve.action,self.lorentzcurve.action,self.voigtcurve.action,self.stepcurve.action,self.doorcurve.action

    def set_fixed(self):
        self.trackchange = False
        item = self.item(self.int1, 4)
        item.setText(_fromUtf8("Fixed"))
        entry = str(self.item(self.int1, 0).text())
        self.extendedparams.partry[entry].vary = False
        self.extendedparams.partry[entry].expr = None
        # nothing to do with self.extendedparams.partry[entry].express, we keep it in memory
        self.trackchange = True

    def set_free(self):
        self.trackchange = False
        item = self.item(self.int1, 4)
        item.setText(_fromUtf8("Free"))
        entry = str(self.item(self.int1, 0).text())
        self.extendedparams.partry[entry].vary = True
        self.extendedparams.partry[entry].expr = None
        # nothing to do with self.extendedparams.partry[entry].express, we keep it in memory
        self.trackchange = True

    def use_expr(self):
        self.trackchange = False
        item = self.item(self.int1, 4)
        item.setText(_fromUtf8("Expr"))
        entry = str(self.item(self.int1, 0).text())

        txt = str(self.item(self.int1, 7).text())
        self.extendedparams.partry[
            entry
        ].express = txt  # in principle, should have been already set

        if len(txt) != 0:  # on met l'expression indiquee
            self.extendedparams.partry[entry].expr = txt
            self.extendedparams.partry[entry].vary = False

        self.trackchange = True

    def inverse_restrains(self):
        # inverse les contraintes: les parametres libres deviennent fixes et vice-versa
        self.trackchange = False
        for int1 in range(self.rowCount()):
            item = self.item(int1, 4)
            entry = str(self.item(int1, 0).text())
            vary = self.extendedparams.partry[entry].vary
            if vary:
                self.extendedparams.partry[entry].vary = False
                item.setText(_fromUtf8("Fixed"))
            else:
                self.extendedparams.partry[entry].vary = True
                item.setText(_fromUtf8("Free"))
        self.trackchange = True

    def set_wholeastry(self):
        self.trackchange = False
        self.extendedparams.setwholefitastry()
        self.initparams(settry=True, setopt=True)
        self.trackchange = True

    def set_as_try(self):
        self.trackchange = False
        item = self.item(self.int1, 1)
        entry = str(self.item(self.int1, 0).text())
        if entry in self.extendedparams.paropt:
            self.extendedparams.partry[entry].value = self.extendedparams.paropt[
                entry
            ].value
            item.setText("%f" % (self.extendedparams.paropt[entry].value))
            self.drawtry()
        self.trackchange = True

    def drawtrypart(self):
        pass

    def showfitpart(self):
        pass

    def startfit(self):
        self.extendedparams.startfit()
        self.isfitted = True
        self.issaved = False
        self.drawopt()
        self.initparams(settry=False, setopt=True)
        self.emit(self.SIG_FIT_DONE)

    def drawtry(self):
        xmin, xmax = self.plot.get_axis_limits("bottom")
        xtry = np.linspace(xmin, xmax, num=10000)
        ytry = model(
            self.extendedparams.partry,
            xtry,
            (self.extendedparams.degpol, self.extendedparams.curveslist),
        )
        self.plot.curvetry.set_data(xtry, ytry)
        self.plot.show_items((self.plot.curveexp, self.plot.curvetry))
        self.plot.hide_items((self.plot.curveopt,))

    def drawopt(self):
        xmin, xmax = self.plot.get_axis_limits("bottom")
        xopt = np.linspace(xmin, xmax, num=10000)
        yopt = model(
            self.extendedparams.paropt,
            xopt,
            (self.extendedparams.degpol, self.extendedparams.curveslist),
        )
        self.plot.curveopt.set_data(xopt, yopt)
        self.plot.show_items((self.plot.curveexp, self.plot.curveopt))
        self.plot.hide_items((self.plot.curvetry,))

    def save_fit(self):
        # sauvegarde des donnees
        if self.savename is None:
            self.setfilename()  # definit le nom de fichier

        if self.savename is not None:
            # dans le cas contraire, c'est qu'on a annule a l'etape precedente
            self.fic = open(self.savename, "a")

            # on ecrit le nom des variables
            lig = "Scan "
            # on ecrit le nom des variables passes en tags
            for kk in self.tags:
                lig = lig + kk[0] + " "

            # on ecrit la valeur des integrales brutes
            if self.saveint:
                lig = lig + "int_tot "

                for i in range(len(self.extendedparams.curveslist)):
                    lig = lig + "int_%d " % (i + 1)

            # on ecrit le nom des variables de fit
            for kk in self.extendedparams.partry:
                lig = lig + kk + " "
            lig = lig + " type \n"
            # print lig
            self.fic.write(lig)

            # on ecrit la valeur des variables
            title = self.scantitle

            lig = title + " "
            # on ecrit la valeur des variables passes en tags
            for kk in self.tags:
                lig = lig + kk[1] + " "

            # on ecrit la valeur des integrales brutes
            if self.saveint:
                self.ints = integral(
                    self.extendedparams.paropt, self.extendedparams.curveslist
                )
                lig = lig + "%g " % (self.ints[0])
                for i in range(len(self.extendedparams.curveslist)):
                    if self.ints[i + 1] is None:
                        lig = lig + "None "
                    else:
                        lig = lig + "%g " % (self.ints[i + 1])

            for kk in self.extendedparams.partry:
                lig = lig + "%g " % (self.extendedparams.paropt[kk].value)
            typ = ""
            for func in self.extendedparams.curveslist:
                typ = typ + func[0]  # on garde en abreviation le 1er caractere "
            lig = lig + typ + "\n"
            # print lig
            self.fic.write(lig)

            # on ecrit une ligne avec les erreurs
            lig = "sigma "
            if self.saveint:
                # ici on n'a pas integre l'evaluation des integrale car elle necessite
                # la diagonalisation de la matrice des covariances.
                lig = lig + "0. "
                for i in range(len(self.extendedparams.curveslist)):
                    lig = lig + "0. "
            for kk in self.extendedparams.partry:
                if self.extendedparams.paropt[kk].stderr is None:
                    lig = lig + "None "
                else:
                    lig = lig + "%g " % (self.extendedparams.paropt[kk].stderr)
            lig = lig + typ + "\n"
            self.fic.write(lig)
            self.fic.close()

            # on sauve la figure de fit
            ific = 1

            figshortname = self.scantitle + "_%.3d.png" % (ific)
            figname = path.join(self.cwd, figshortname)
            while path.isfile(figname):
                ific = ific + 1
                figshortname = self.scantitle + "_%.3d.png" % (ific)
                figname = path.join(self.cwd, figshortname)
            print(figname, " sauve")
            self.plot.save_widget(figname)
            self.issaved = True
            if self.updatestart:
                self.set_wholeastry()

            self.emit(self.SIG_FIT_DONE)

    def setfilename(self):
        self.savename = str(
            QFileDialog.getSaveFileName(
                None, "Save fit parameters", self.cwd, filter="*.txt"
            )
        )
        if len(self.savename) == 0:
            self.savename = None
        else:
            self.cwd = path.dirname(path.realpath(self.savename))

    def rangemove(self, shape):
        # on recalcule le masque
        items = self.plot.get_items(z_sorted=False, item_type=IShapeItemType)
        ranges = list()
        for item in items:
            ranges.append(item.get_range())

        if shape not in items:
            ranges.append(shape.get_range())
            # alors il s'agit d'une nouvelle XRangeSelection
        self.extendedparams.mask(ranges)

    def rangeremove(self, shape):
        self.setrange()

    def setrange(self):
        # on recalcule le masque
        items = self.plot.get_items(z_sorted=False, item_type=IShapeItemType)
        ranges = list()
        for item in items:
            ranges.append(item.get_range())
        self.extendedparams.mask(ranges)

    def shiftguess(self, shift):
        # on shifte toutes les positions guess des courbes
        # attention, si les positions sont fixes, ca decale les courbes quand meme
        # si on ne veut pas decaler, alors il faut mettre la valeur dans Expression
        for int1 in range(self.rowCount()):
            txt = str(self.item(int1, 1).text())
            entry = str(self.item(int1, 0).text())

            if "pos" in entry[:3]:  # alors c'est un parametre de position a shifter
                val = float(txt) + shift
                self.item(int1, 1).setText("%f" % (val))


class ExtendedParams:
    # dictionnaire de la liste des courbes pour le fit
    # utilise pour comprendre les Parameters de lmfit
    # ne pilote jamais le tracage des courbes et l'ecriture dans le tableau
    def __init__(self):
        self.degpol = -1  # degre du polynome de fond
        self.ncurves = 0  # nombre de courbes
        self.curveslist = (
            list()
        )  # liste d'un dictionnaire comprenant type de courbe et nom des parametres
        self.partry = Parameters()  # parametres d'essai
        self.paropt = Parameters()  # parametres optimises
        self.xexp = np.zeros((1,))
        self.yexp = np.zeros((1,))
        self.tags = []

    def printparams(self):
        print("printparams")
        for key in self.partry.keys():
            print(key, self.partry[key].value, self.partry[key].vary)

    def addbackground(self):
        k = self.degpol + 1
        kstr = "%d" % (k)
        entry = "pol" + kstr
        self.degpol = k
        self.partry.add(entry, value=0.00)
        self.partry[
            entry
        ].number = (
            0  # on rajoute une reference au nombre de la courbe (0 pour le polynome)
        )
        self.partry[
            entry
        ].express = ""  # on ajoute a la classe une variable 'express' qui est la valeur de l'expression pas forcement appliquee (pour la garder en memoire)

    def removebackground(self):
        k = self.degpol
        kstr = "%d" % (k)
        entry = "pol" + kstr
        del self.partry[entry]
        self.degpol = k - 1

    def addcurve(self, curvename="gausian", paramvalues=None):
        # ajoute une courbe dans la liste parametres
        self.curveslist.append(curvename)
        k = self.ncurves + 1
        if paramvalues is None:
            # on prend les valeurs par defaut
            for ent, val, vmin, vmax in curveparams[curvename]:
                entry = ent + "%d" % (k)
                self.partry.add(entry, value=val)
                self.partry[entry].express = ""
                self.partry[entry].number = k
                if vmin is not None:
                    self.partry[entry].min = vmin
                if vmax is not None:
                    self.partry[entry].max = vmax

        else:
            # on prend les valeurs de paramvalues
            for cp, val in zip(curveparams[curvename], paramvalues):
                entry = cp[0] + "%d" % (k)
                self.partry.add(entry, value=val)
                self.partry[entry].express = ""
                self.partry[entry].number = k
                vmin = cp[2]
                vmax = cp[3]
                if vmin is not None:
                    self.partry[entry].min = vmin
                if vmax is not None:
                    self.partry[entry].max = vmax

        self.ncurves = k

    def freezecurve(self, k):
        # freeze tout les parametres associes a la courbe
        curvename = self.curveslist[k - 1]
        print(curvename)
        for cp in curveparams[curvename]:
            ent = cp[0]
            entry = ent + "%d" % (k)
            self.partry[entry].vary = False
            print(entry, "Fixed")

    def releasecurve(self, k):
        # freeze tout les parametres associes a la courbe
        curvename = self.curveslist[k - 1]
        print(curvename)
        for cp in curveparams[curvename]:
            ent = cp[0]
            entry = ent + "%d" % (k)
            self.partry[entry].vary = True
            print(entry, "Free")

    def removecurve(self, k):
        ncurve = len(self.curveslist)
        curvename = self.curveslist.pop(k - 1)

        for cp in curveparams[curvename]:
            ent = cp[0]
            entry = ent + "%d" % (k)
            del self.partry[entry]

        # reorganisation des courbes
        for k2 in range(k, ncurve):
            curvename = self.curveslist[k2 - 1]
            for cp in curveparams[curvename]:
                ent = cp[0]
                entry = ent + "%d" % (k2 + 1)
                entry2 = ent + "%d" % (k2)  # on redescend les courbes d'une unite
                self.partry.add(
                    entry2,
                    value=self.partry[entry].value,
                    vary=self.partry[entry].vary,
                    min=self.partry[entry].min,
                    max=self.partry[entry].max,
                    expr=self.partry[entry].expr,
                )
                self.partry[entry2].number = k2
                self.partry[entry2].express = ""
                del self.partry[entry]

        self.ncurves = self.ncurves - 1

    def changecurve(self, paramvalues=None, icurve=-1):
        # icurve est le numero de la liste des courbes, a partir de 0
        curvename = self.curveslist[icurve]
        k = icurve + 1
        if icurve == -1:
            k = len(self.curveslist)

        if paramvalues is None:
            # on prend les valeurs par defaut
            for cp in curveparams[curvename]:
                ent = cp[0]
                val = cp[1]
                entry = ent + "%d" % (k)
                self.partry[entry].value = val
                # print entry,self.partry[entry]
        else:
            # on prend les valeurs de paramvalues
            for cp, val in zip(curveparams[curvename], paramvalues):
                entry = cp[0] + "%d" % (k)
                self.partry[entry].value = val

    def setvalexp(self, xexp, yexp):
        self.xexp = xexp
        self.yexp = yexp
        self.xexp0 = xexp
        self.yexp0 = yexp

    def mask(self, ranges):
        xexp = np.ma.array(self.xexp0)
        for rang in ranges:
            xexp = np.ma.masked_inside(xexp, rang[0], rang[1])
        mask = np.ma.getmaskarray(xexp)

        yexp = np.ma.array(self.yexp0, mask=mask)
        self.xexp = xexp.compressed()
        self.yexp = yexp.compressed()

    def startfit(self):
        # self.paropt=copy.deepcopy(self.partry)
        """
        print "self.paropt:",self.paropt
        print "self.degpol:",self.degpol
        print "self.curveslist:",self.curveslist
        print "self.xexp",self.xexp
        print "self.yexp",self.yexp
        """
        # modification of lmfit

        result = minimize(
            residual,
            self.partry,
            args=(self.xexp, self.yexp, (self.degpol, self.curveslist)),
        )
        self.paropt = result.params
        result.params.pretty_print()
        """
        for kk in self.partry:
            pkt=self.partry[kk].value
            pkmi=self.partry[kk].min
            pkma=self.partry[kk].max
            pko=self.paropt[kk].value
            pkv=self.paropt[kk].vary
            pke=self.paropt[kk].expr
            print kk,pkt,pko,pkv,pkmi,pkma,pke
        """
        self.integrate()
        print(integral(self.paropt, self.curveslist))

    def integrate(self):
        # calcule l'integrale brute en enlevant le fond
        self.int1 = 0.0
        self.int2 = 0.0
        ncurv = len(self.curveslist)
        if ncurv > 0:
            integraltot = simps(self.yexp, self.xexp)
            # integraltot=np.sum(self.yexp)
            ybg = model(self.paropt, self.xexp, (self.degpol, list()))
            # self.int1=integraltot-np.sum(ybg)
            self.int1 = integraltot - simps(ybg, self.xexp)
            print("integrale brute fond soustrait:", self.int1)
            # calcule l'integrale brute en enlevant toutes les courbes sauf la premiere
            self.parpart = copy.deepcopy(self.paropt)
            self.parpart["int1"].value = 0.0
            ybg = model(self.parpart, self.xexp, (self.degpol, self.curveslist))
            # self.int2=integraltot-np.sum(ybg)
            self.int2 = integraltot - simps(ybg, self.xexp)
            print("integrale brute fond + courbes >1 soustraits:", self.int2)
            yth = model(self.paropt, self.xexp, (-1, self.curveslist))
            self.int3 = simps(yth, self.xexp)
            print("integrale numerique courbes simulees:", self.int3)

    def setwholefitastry(self):
        for kk in self.partry:
            if kk in self.paropt:
                self.partry[kk].value = self.paropt[kk].value

    def updatetags(self, tags):
        # on supprime les entrees precedentes:
        for tag in self.tags:
            if tag[0] in tags:  # le tag existe encore, on update sa valeur
                self.partry[tag[0]].value = tag[1]
            else:
                del self.partry[tag[0]]  # le tag a disparu, on le supprime

        for tag in tags:
            if tag[0] not in tags:  # le tag n'existe pas encore, on l'ajoute
                self.partry.add(tag[0], value=tag[1], vary=False)
        self.tags = tags


class MaskTool(BaseCursorTool):
    TITLE = _("Mask data")
    ICON = "xrange.png"
    SWITCH_TO_DEFAULT_TOOL = True

    def __init__(
        self, manager, toolbar_id=DefaultToolbarID, title=None, icon=None, tip=None
    ):
        super(MaskTool, self).__init__(
            manager, toolbar_id, title=title, icon=icon, tip=tip
        )
        self._last_item = None

    def get_last_item(self):
        if self._last_item is not None:
            return self._last_item()

    def create_shape(self):
        return XRangeSelection(0, 0)

    def move(self, filter, event):
        super(MaskTool, self).move(filter, event)

    def end_move(self, filter, event):
        super(MaskTool, self).end_move(filter, event)

    def get_associated_item(self, plot):
        items = plot.get_selected_items(item_type=ICurveItemType)
        if len(items) == 1:
            self._last_item = weakref.ref(items[0])
        return self.get_last_item()

    def update_status(self, plot):
        pass
        # item = self.get_associated_item(plot)
        # self.action.setEnabled(item is not None)


class FitTool(CommandTool):
    def __init__(self, manager, toolbar_id=DefaultToolbarID):
        CommandTool.__init__(
            self,
            manager,
            _("Fit"),
            icon=gausspath,
            tip=_("Add curve for fit"),
            toolbar_id=toolbar_id,
        )
        self.manager = manager

    def create_action_menu(self, manager):
        # Create and return menu for the tool's action
        """
        self.gausscurve = manager.add_tool(RectangularActionToolCX,actiongauss, toolbar_id=None, icon="rectangle.png",title = "Gauss")
        self.lorentzcurve = manager.add_tool(RectangularActionToolCX,actionlorentz, toolbar_id=None, icon="rectangle.png",title = "Lorentz")
        self.voigtcurve = manager.add_tool(RectangularActionToolCX,actionvoigt, toolbar_id=None, icon="rectangle.png",title = "Voigt")
        self.stepcurve = manager.add_tool(RectangularActionToolCXY ,actionstep, toolbar_id=None, icon="rectangle.png",title = "Step")
        self.doorcurve = manager.add_tool(RectangularActionToolCX,actiondoor, toolbar_id=None, icon="rectangle.png",title = "Door")
        """
        gaussaction = QAction(QIcon(gausspath), "Gaussian", self)
        self.connect(gaussaction, SIGNAL(_fromUtf8("triggered()")), self.drawnewgauss)

        lorentzaction = QAction(QIcon(gausspath), "Lorentzian", self)
        self.connect(
            lorentzaction, SIGNAL(_fromUtf8("triggered()")), self.drawnewlorentz
        )

        voigtaction = QAction(QIcon(gausspath), "Voigt", self)
        self.connect(voigtaction, SIGNAL(_fromUtf8("triggered()")), self.drawnewvoigt)

        stepaction = QAction(QIcon(steppath), "Step", self)
        self.connect(stepaction, SIGNAL(_fromUtf8("triggered()")), self.drawnewstep)

        dooraction = QAction(QIcon(doorpath), "Door", self)
        self.connect(dooraction, SIGNAL(_fromUtf8("triggered()")), self.drawnewdoor)

        gaussderaction = QAction(QIcon(gaussderpath), "Gaussian_derivative", self)
        self.connect(
            gaussderaction, SIGNAL(_fromUtf8("triggered()")), self.drawnewgaussder
        )

        lorentzderaction = QAction(QIcon(gaussderpath), "Lorentz_derivative", self)
        self.connect(
            lorentzderaction, SIGNAL(_fromUtf8("triggered()")), self.drawnewlorentzder
        )

        gaussloraction = QAction(QIcon(gausslorpath), "Gaussian_lorentzian", self)
        self.connect(
            gaussloraction, SIGNAL(_fromUtf8("triggered()")), self.drawnewgausslor
        )

        skewaction = QAction(QIcon(skewpath), "Skew", self)
        self.connect(skewaction, SIGNAL(_fromUtf8("triggered()")), self.drawnewskew)

        resetaction = QAction(QIcon(eraserpath), "Reset", self)
        self.connect(resetaction, SIGNAL(_fromUtf8("triggered()")), self.reset)

        menu = QMenu()
        # add_actions(menu, (self.gausscurve.action,self.lorentzcurve.action,self.voigtcurve.action,self.stepcurve.action,self.doorcurve.action,resetaction))
        add_actions(
            menu,
            (
                gaussaction,
                lorentzaction,
                voigtaction,
                stepaction,
                dooraction,
                gaussderaction,
                lorentzderaction,
                gaussloraction,
                skewaction,
                resetaction,
            ),
        )

        self.gausstool = self.manager.add_tool(
            RectangularActionToolCX,
            actiongauss,
            toolbar_id=None,
            icon="rectangle.png",
            title="Gauss",
        )
        self.lorentztool = self.manager.add_tool(
            RectangularActionToolCX,
            actionlorentz,
            toolbar_id=None,
            icon="rectangle.png",
            title="Lorentz",
        )
        self.voigttool = self.manager.add_tool(
            RectangularActionToolCX,
            actionvoigt,
            toolbar_id=None,
            icon="rectangle.png",
            title="Voigt",
        )
        self.steptool = self.manager.add_tool(
            RectangularActionToolCXY,
            actionstep,
            toolbar_id=None,
            icon="rectangle.png",
            title="Step",
        )
        self.doortool = self.manager.add_tool(
            RectangularActionToolCX,
            actiondoor,
            toolbar_id=None,
            icon="rectangle.png",
            title="Door",
        )
        self.gaussdertool = self.manager.add_tool(
            RectangularActionToolCX,
            actiongaussder,
            toolbar_id=None,
            icon="rectangle.png",
            title="GaussDer",
        )
        self.lorentzdertool = self.manager.add_tool(
            RectangularActionToolCX,
            actionlorentzder,
            toolbar_id=None,
            icon="rectangle.png",
            title="LorentzDer",
        )
        self.gausslortool = self.manager.add_tool(
            RectangularActionToolCX,
            actiongausslor,
            toolbar_id=None,
            icon="rectangle.png",
            title="GaussLor",
        )
        self.skewtool = self.manager.add_tool(
            RectangularActionToolCX,
            actionskew,
            toolbar_id=None,
            icon="rectangle.png",
            title="Skew",
        )
        # print "g",self.gausstool
        # print "l",self.lorentztool
        # print "v",self.voigttool
        # print "s",self.steptool
        # print "d",self.doortool

        self.action.setMenu(menu)
        return menu

    def reset(self):
        plot = self.get_active_plot()
        self.manager.activate_default_tool()
        plot.tabval.reset()
        plot.tabval.extendedparams.setvalexp(plot.xexp, plot.yexp)
        plot.tabval.drawtry()

    def drawnewgauss(self):
        # print "drawnewgauss"
        plot = self.get_active_plot()
        plot.tabval.setactivecurve("new")
        self.gausstool.activate()

    def drawnewlorentz(self):
        # print "drawnewlorentz"
        plot = self.get_active_plot()
        plot.tabval.setactivecurve("new")
        self.lorentztool.activate()

    def drawnewvoigt(self):
        # print "drawnewvoigt"
        plot = self.get_active_plot()
        plot.tabval.setactivecurve("new")
        self.voigttool.activate()

    def drawnewstep(self):
        # print "drawnewstep"
        plot = self.get_active_plot()
        plot.tabval.setactivecurve("new")
        self.steptool.activate()

    def drawnewdoor(self):
        # print "drawnewdoor"
        plot = self.get_active_plot()
        plot.tabval.setactivecurve("new")
        self.doortool.activate()

    def drawnewgaussder(self):
        # print "drawnewgauss"
        plot = self.get_active_plot()
        plot.tabval.setactivecurve("new")
        self.gaussdertool.activate()

    def drawnewlorentzder(self):
        # print "drawnewgauss"
        plot = self.get_active_plot()
        plot.tabval.setactivecurve("new")
        self.lorentzdertool.activate()

    def drawnewgausslor(self):
        # print "drawnewgauss"
        plot = self.get_active_plot()
        plot.tabval.setactivecurve("new")
        self.gausslortool.activate()

    def drawnewskew(self):
        # print "drawnewgauss"
        plot = self.get_active_plot()
        plot.tabval.setactivecurve("new")
        self.skewtool.activate()

    def handle_shape(self, shape, inside):
        shape.set_style("plot", "shape/mask")
        shape.set_private(True)
        plot = self.get_active_plot()
        plot.set_active_item(shape)
        self._mask_shapes[plot] += [(shape, inside)]

    def deactivate(self):
        """Deactivate tools"""
        print("deactivate")
        self.gausscurve.deactivate()
        self.lorentzcurve.deactivate()
        self.voigtcurve.deactivate()
        self.stepcurve.deactivate()
        self.doorcurve.deactivate()
        self.gaussdercurve.deactivate()
        self.lorentzdercurve.deactivate()
        print("deactivate done")


class RunTool(CommandTool):
    def __init__(self, manager, toolbar_id=DefaultToolbarID):
        """
        CommandTool.__init__(self, manager, _("Run"),icon=QIcon("Run.png"),
                                            tip=_("Perform fit"),
                                            toolbar_id=toolbar_id)
        """

        CommandTool.__init__(
            self,
            manager,
            _("Run"),
            icon=runpath,
            tip=_("Perform fit"),
            toolbar_id=toolbar_id,
        )

    def activate_command(self, plot, checked):
        plot = self.get_active_plot()
        plot.tabval.startfit()
        # Activate tool
        pass


class YFullRangeTool(CommandTool):
    def __init__(self, manager, toolbar_id=DefaultToolbarID):
        CommandTool.__init__(
            self,
            manager,
            _("Full Range"),
            icon=yfullrangepath,
            tip=_("Set Y Full Range"),
            toolbar_id=toolbar_id,
        )

    def activate_command(self, plot, checked):
        plot = self.get_active_plot()
        vmin = np.amin(plot.tabval.extendedparams.yexp)
        vmax = np.amax(plot.tabval.extendedparams.yexp)
        plot.set_axis_limits("left", vmin, vmax)
        plot.replot()


class PrefTool(CommandTool):
    def __init__(self, manager, toolbar_id=DefaultToolbarID):
        CommandTool.__init__(
            self,
            manager,
            _("Run"),
            icon="settings.png",
            tip=_("Preferences"),
            toolbar_id=toolbar_id,
        )
        self.manager = manager

    def create_action_menu(self, manager):
        # Create and return menu for the tool's action
        self.saveprefaction = QAction("Save as...", self)
        self.connect(
            self.saveprefaction, SIGNAL(_fromUtf8("triggered()")), self.savepref
        )

        self.updatestartaction = manager.create_action(
            _("Start with last fit"), toggled=self.updatestart
        )
        self.updatestartaction.setChecked(False)

        self.showtagsaction = QAction("Show tags", self)
        self.connect(
            self.showtagsaction, SIGNAL(_fromUtf8("triggered()")), self.showtags
        )

        self.scaleprefaction = QAction("Autoscale", self)
        self.scaleprefaction.setCheckable(True)
        self.scaleprefaction.setChecked(True)

        menu = QMenu()
        # add_actions(menu, (self.gausscurve.action,self.lorentzcurve.action,self.voigtcurve.action,self.stepcurve.action,self.doorcurve.action,resetaction))
        add_actions(
            menu,
            (
                self.saveprefaction,
                self.updatestartaction,
                self.showtagsaction,
                self.scaleprefaction,
            ),
        )

        self.action.setMenu(menu)
        return menu

    def savepref(self):
        # print "drawnewlorentz"
        plot = self.get_active_plot()
        plot.tabval.setfilename()

    def showtags(self):
        # print "drawnewlorentz"
        plot = self.get_active_plot()
        tags = plot.tabval.extendedparams.tags
        # print 'tags'
        TagsWindow(tags)

    def updatestart(self):
        # print "drawnewlorentz"
        plot = self.get_active_plot()
        if self.updatestartaction.isChecked():
            plot.tabval.updatestart = True
        else:
            plot.tabval.updatestart = False

    def setupdatestart(self, value):
        plot = self.get_active_plot()
        self.updatestartaction.setChecked(value)
        plot.tabval.updatestart = value


class SaveFitTool(CommandTool):
    def __init__(self, manager, toolbar_id=DefaultToolbarID):
        super(SaveFitTool, self).__init__(
            manager,
            _("Save fit result"),
            get_std_icon("DialogSaveButton", 16),
            toolbar_id=toolbar_id,
        )

    def activate_command(self, plot, checked):
        if plot.tabval.isfitted:
            plot.tabval.save_fit()


class RectangularSelectionHandlerCX(RectangularSelectionHandler):
    # on utilise la classe heritee dont on surcharge la methode move

    def move(self, filter, event):
        """methode surchargee par la classe """
        sympos = QPoint(2 * self.start.x() - event.pos().x(), self.start.y())
        self.shape.move_local_point_to(self.shape_h0, sympos)
        self.shape.move_local_point_to(self.shape_h1, event.pos())
        self.move_action(filter, event)
        filter.plot.replot()


class RectangularActionToolCX(RectangularActionTool):
    # outil de tracage de rectangles centre en x
    # on utilise la classe heritee dont on surcharge la methode setup_filter

    def setup_filter(
        self, baseplot
    ):  # partie utilisee pendant le mouvement a la souris
        # utilise a l'initialisation de la toolbar
        # print "setup_filter"
        filter = baseplot.filter
        start_state = filter.new_state()
        handler = RectangularSelectionHandlerCX(
            filter, Qt.LeftButton, start_state=start_state  # gestionnaire du filtre
        )
        shape, h0, h1 = self.get_shape()
        shape.pen.setColor(QColor("#00bfff"))

        handler.set_shape(
            shape, h0, h1, self.setup_shape, avoid_null_shape=self.AVOID_NULL_SHAPE
        )
        self.connect(handler, SIG_END_RECT, self.end_rect)
        # self.connect(handler, SIG_CLICK_EVENT, self.start)   #a definir aussi dans RectangularSelectionHandler2
        return setup_standard_tool_filter(filter, start_state)

    def activate(self):
        """Activate tool"""

        # print "commande active",self
        for baseplot, start_state in self.start_state.items():
            baseplot.filter.set_state(start_state, None)
        self.action.setChecked(True)
        self.manager.set_active_tool(self)
        # plot = self.get_active_plot()
        # plot.newcurve=True

    def deactivate(self):
        """Deactivate tool"""
        # print "commande desactivee",self
        self.action.setChecked(False)


class RectangularSelectionHandlerCXY(RectangularSelectionHandler):
    # on utilise la classe heritee dont on surcharge la methode move

    def move(self, filter, event):
        """methode surchargee par la classe """
        sympos = QPoint(
            2 * self.start.x() - event.pos().x(), 2 * self.start.y() - event.pos().y()
        )
        self.shape.move_local_point_to(self.shape_h0, sympos)
        self.shape.move_local_point_to(self.shape_h1, event.pos())
        self.move_action(filter, event)
        filter.plot.replot()


class RectangularActionToolCXY(RectangularActionTool):
    # outil de tracage de rectangles centre en x
    # on utilise la classe heritee dont on surcharge la methode setup_filter

    def setup_filter(
        self, baseplot
    ):  # partie utilisee pendant le mouvement a la souris
        # utilise a l'initialisation de la toolbar
        # print "setup_filter"
        filter = baseplot.filter
        start_state = filter.new_state()
        handler = RectangularSelectionHandlerCXY(
            filter, Qt.LeftButton, start_state=start_state  # gestionnaire du filtre
        )
        shape, h0, h1 = self.get_shape()
        shape.pen.setColor(QColor("#00bfff"))

        handler.set_shape(
            shape, h0, h1, self.setup_shape, avoid_null_shape=self.AVOID_NULL_SHAPE
        )
        self.connect(handler, SIG_END_RECT, self.end_rect)
        # self.connect(handler, SIG_CLICK_EVENT, self.start)   #a definir aussi dans RectangularSelectionHandler2
        return setup_standard_tool_filter(filter, start_state)

    def activate(self):
        """Activate tool"""

        # print "commande active",self
        for baseplot, start_state in self.start_state.items():
            baseplot.filter.set_state(start_state, None)
        self.action.setChecked(True)
        self.manager.set_active_tool(self)
        # plot = self.get_active_plot()
        # plot.newcurve=True

    def deactivate(self):
        """Deactivate tool"""
        # print "commande desactivee",self
        self.action.setChecked(False)


def get_xiwb_cx(plot, p0, p1):
    # return position, intensity,fwhm and background  from the rectangular tool cx
    ax, ay = plot.get_axis_id("bottom"), plot.get_axis_id("left")
    x1, y1 = plot.invTransform(ax, p0.x()), plot.invTransform(ay, p0.y())
    x2, y2 = plot.invTransform(ax, p1.x()), plot.invTransform(ay, p1.y())
    return x1, y1 - y2, 2.0 * (x2 - x1), y2


def get_xiwb_cxy(plot, p0, p1):
    # return position, intensity,width and background  from the rectangular tool cxy
    ax, ay = plot.get_axis_id("bottom"), plot.get_axis_id("left")
    x1, y1 = plot.invTransform(ax, p0.x()), plot.invTransform(ay, p0.y())
    x2, y2 = plot.invTransform(ax, p1.x()), plot.invTransform(ay, p1.y())
    return x1, 2 * (y1 - y2), 2.0 * (x2 - x1), y2


def actiongauss(plot, p0, p1):
    x0try, itry, wtry, btry = get_xiwb_cx(plot, p0, p1)
    plot.tabval.set_cv("gaussian", (itry, x0try, wtry), bg=btry)

    # lim=plot.get_axis_limits("bottom")
    # xtry=np.linspace(lim[0],lim[1], num=100)
    # ytry=gauss(xtry,[itry,x0try,wtry])+btry

    """
    #pour rajouter un polygone
    points=np.concatenate((xtry,ytry))
    points=points.reshape((2,100))
    points=np.transpose(points)
    testpolygon=PolygonShape(points=points)
    testpolygon.set_selectable(False)
    plot.add_item(testpolygon)
    """


def actionlorentz(plot, p0, p1):
    x0try, itry, wtry, btry = get_xiwb_cx(plot, p0, p1)
    plot.tabval.set_cv("lorentzian", (itry, x0try, wtry), bg=btry)


def actionvoigt(plot, p0, p1):
    # trace une pseudo-voigt avec poids egal sur les pics lorentz et gauss
    x0try, itry, wtry, btry = get_xiwb_cx(plot, p0, p1)
    plot.tabval.set_cv("voigt", (itry, x0try, wtry, 0.5), bg=btry)


def actionstep(plot, p0, p1):
    x0try, itry, wtry, btry = get_xiwb_cxy(plot, p0, p1)
    plot.tabval.set_cv("step", (itry, x0try, wtry), bg=btry)


def actiondoor(plot, p0, p1):
    x0try, itry, wtry, btry = get_xiwb_cx(plot, p0, p1)
    plot.tabval.set_cv("door", (itry, x0try, wtry, wtry / 4.0), bg=btry)


def actiongaussder(plot, p0, p1):
    x0try, itry, wtry, btry = get_xiwb_cx(plot, p0, p1)
    plot.tabval.set_cv("gaussian_derivative", (itry, x0try, wtry), bg=btry + itry / 2.0)


def actiongausslor(plot, p0, p1):
    x0try, itry, wtry, btry = get_xiwb_cx(plot, p0, p1)
    plot.tabval.set_cv("gaussian_lorentzian", (itry, x0try, wtry, 0.1), bg=btry)


def actionlorentzder(plot, p0, p1):
    x0try, itry, wtry, btry = get_xiwb_cx(plot, p0, p1)
    plot.tabval.set_cv("gaussian_derivative", (itry, x0try, wtry), bg=btry + itry / 2.0)


def actionskew(plot, p0, p1):
    x0try, itry, wtry, btry = get_xiwb_cx(plot, p0, p1)
    plot.tabval.set_cv("skew", (itry, x0try, wtry, 0.0), bg=btry)


class Ui_FitWindow(QMainWindow):
    def setupUi(self):
        self.setObjectName(_fromUtf8("MainWindow"))

        self.setWindowTitle("Fit window")

        self.savename = None

        # on utilise un splitter pour separer la fenetre en deux, ce sera le central widget
        self.splitter = QSplitter(Qt.Vertical)

        # widget pour afficher les courbes
        self.cv = CurveWidget(parent=self, show_itemlist=True)

        # widget pour afficher les donnees du fit , gerer les entrees de parametres
        self.tabval = FitTable(parent=self)

        self.splitter.addWidget(self.cv)
        self.splitter.addWidget(self.tabval)

        self.setCentralWidget(self.splitter)

        # on utilise le plot manager de guiqwt
        self.manager = PlotManager(self)
        self.plot = self.cv.get_plot()

        # on associe les references croisees au tableau des donnees et plot
        self.plot.tabval = self.tabval
        self.tabval.plot = self.plot

        # attribut pour dire qu'on a demande une nouvelle courbe qui n'est pas encore ajustee a la souris
        self.plot.newcurve = False

        # default empty curves
        z = np.empty((0,))
        self.empty = True
        self.plot.curveexp = make.curve(
            z,
            z,
            marker="Diamond",
            markerfacecolor="r",
            markeredgecolor="r",
            markersize=4,
            linestyle="NoPen",
            title="experiment",
        )
        self.plot.curvetry = make.curve(z, z, color="g", title="estimate")
        self.plot.curveopt = make.curve(z, z, color="k", title="fit result")

        self.plot.curveexp.set_readonly(True)
        self.plot.curvetry.set_readonly(True)
        self.plot.curveopt.set_readonly(True)

        self.plot.add_item(self.plot.curveexp)
        self.plot.add_item(self.plot.curvetry)
        self.plot.add_item(self.plot.curveopt)

        self.plot.hide_items(
            (self.plot.curveexp, self.plot.curvetry, self.plot.curveopt)
        )

        self.manager.add_plot(self.plot)

        self.connect(self.plot, SIG_RANGE_CHANGED, self.tabval.rangemove)
        self.connect(self.plot, SIG_ITEM_REMOVED, self.tabval.rangeremove)
        # self.connect(self.tabval, SIG_FIT_DONE, self.coucou)

        # ---Add toolbar and register manager tools
        toolbar = self.addToolBar("tools")

        self.manager.add_toolbar(toolbar, id(toolbar))

        # self.manager.register_all_curve_tools()
        self.manager.register_standard_tools()
        self.manager.register_other_tools()

        self.oft = self.manager.add_tool(OpenFileTool)
        self.oft.formats = " *.txt *.dat"
        self.title = "Open data file"
        self.oft.connect(self.oft, SIGNAL("openfile(QString*)"), self.open_file)

        self.FitTool = self.manager.add_tool(FitTool)  # ok
        self.MaskTool = self.manager.add_tool(MaskTool)
        self.manager.add_tool(RunTool)  # ok
        self.manager.add_tool(SaveFitTool)
        self.manager.add_tool(YFullRangeTool)
        self.Pref_Tool = self.manager.add_tool(PrefTool)
        self.Pref_Tool.setupdatestart(self.tabval.updatestart)
        # redimensionnement de la fenetre principale, les widgets suivent
        # sinon, prennent leur SizeHint (300,400) pour cv par exemple
        self.resize(820, 600)
        self.readfileparams = readfileparams()
        # pour savoir si un fit a ete fait et sauve

    def coucou(self):
        print("coucou")

    def reset(self):
        self.FitTool.reset()

        # testpolygon=PolygonShape(points=[[1,0],[1.1,0.1],[1.2,0.5],[1.4,2]])
        # self.plot.add_item(testpolygon)

    def open_file(self, QString):

        fname = str(QString)
        shortname = path.split(fname)[1]
        fic = open(fname)
        ficlines = fic.readlines()
        readfilewindow(self.readfileparams, ficlines)
        seps = [" ", "\t", ","]
        sep = seps[self.readfileparams.delimiter]
        nf = len(ficlines)
        ix = self.readfileparams.x - 1
        iy = self.readfileparams.y - 1

        if self.readfileparams.title:
            try:
                ni = self.readfileparams.heading + 1
                title = ficlines[self.readfileparams.heading].split(sep)
                print(title)
                xlabel = title[ix]
                ylabel = title[iy]
            except:
                ni = self.readfileparams.heading
                xlabel = "x"
                ylabel = "y"
        else:
            ni = self.readfileparams.heading
            xlabel = "x"
            ylabel = "y"

        x = []
        y = []
        for i in range(ni, nf):
            try:
                ll = ficlines[i].split(sep)
                xx = float(ll[ix])
                yy = float(ll[iy])
                if np.isfinite(xx) and np.isfinite(yy):
                    x.append(xx)
                    y.append(yy)
            except:
                pass

        self.tabval.cwd = path.dirname(path.realpath(fname))
        self.setvalexp(x, y, xlabel=xlabel, ylabel=ylabel, title=shortname)

    def settags(self, tags, saveint=None):
        # tags est une liste de couples de valeurs transmis par l'application qui sollicite grafit pour
        # etre enregistre dans le fichier de sauvegarde

        self.tabval.extendedparams.updatetags(tags)

        # soit saveint n'est pas modifie et rien ne change, soit saveint l'est et on change
        if saveint is not None:
            self.saveint = saveint
            self.tabval.saveint = saveint

            if self.saveint:
                print("on sauve les integrales")
            else:
                print("on ne sauve pas les integrales")

    def setupdatestart(self, value):
        self.tabval.updatestart = value
        self.Pref_Tool.setupdatestart(self.tabval.updatestart)

    def getfitresult(self):
        return [
            self.tabval.extendedparams.paropt,
            integral(
                self.tabval.extendedparams.paropt, self.tabval.extendedparams.curveslist
            ),
        ]

    def setsaveint(self, value=True):
        self.tabval.saveint = value

    def setsavedir(self, cwd):
        self.tabval.cwd = cwd

    def setvalexp(
        self,
        xexp,
        yexp,
        xlabel="x",
        ylabel="y",
        title="scan",
        xlog=False,
        ylog=False,
        tags=[],
        followscans=False,
    ):
        # title est le titre du scan qui est fitte. xlog et ylog determinent si l'echelle est log ou pas
        # infos est une liste de deux strings qui sont enregistrees en plus des parametres du fit
        # lors de la sauvegarde des donnes. Gere par le programme qui appelle grafit.
        i = 1

        if self.tabval.isfitted and self.tabval.issaved is False:
            # avant d'ecraser le fit precedent on demande s'il faut le sauver
            # pose probleme dans le cas ou setvalexp est genere par le deplacement d'une shape
            i = QMessageBox.question(
                self, "save", "do you want to save fit?", "Yes", "No", "Cancel"
            )

        if i == 0:
            self.tabval.save_fit()

        # s'il existe deja un fit, et que followscans=True, alors on decale les valeurs de fit
        # de façon a ce que la position de guess des pics se retrouve au meme endroit par
        # rapport au centre du scan
        if followscans and not self.empty:

            cen0 = (max(self.plot.xexp) + min(self.plot.xexp)) / 2.0
            cen1 = (max(xexp) + min(xexp)) / 2.0
            shift = cen1 - cen0
            print("on shifte de ", shift)
            self.tabval.shiftguess(shift)

        if i < 2:  # si i=2, on ne fait rien

            if self.Pref_Tool.scaleprefaction.isChecked():
                if xlog:
                    self.plot.set_axis_scale("bottom", "log", autoscale=True)
                else:
                    self.plot.set_axis_scale("bottom", "lin", autoscale=True)
                if ylog:
                    self.plot.set_axis_scale("left", "log", autoscale=True)
                else:
                    self.plot.set_axis_scale("left", "lin", autoscale=True)

            self.plot.xexp = xexp
            self.plot.yexp = yexp
            self.tabval.scantitle = title
            self.tabval.extendedparams.setvalexp(xexp, yexp)
            self.settags(tags)

            self.plot.set_titles(title=title, xlabel=xlabel, ylabel=ylabel)
            self.plot.curveexp.set_data(xexp, yexp)

            self.plot.show_items((self.plot.curveexp,))
            self.plot.hide_items((self.plot.curvetry, self.plot.curveopt))

            self.tabval.isfitted = False  # on n'a pas fait de fit sur ce set de donnees

            self.tabval.setrange()
            self.plot.do_autoscale()
            self.empty = False

        # print xexp.shape,yexp.shape


if __name__ == "__main__":

    app = QApplication(sys.argv)

    ui = Ui_FitWindow()
    ui.setupUi()
    ui.xlabel = "x"
    ui.ylabel = "y"
    x = np.arange(0, 3.14, 0.001)
    y = np.sin(x)
    ui.setvalexp(x, y)
    ui.show()

    sys.exit(app.exec_())
